import { SelectorFn } from "./misc";
import { VModel } from "./vmodel";
import { VNode } from "./vnode";
import { InvalidatorSubscription, Invalidator } from "./invalidator";
/**
 * Component Descriptor registry used in DEBUG mode.
 */
export declare const ComponentDescriptorRegistry: Map<string, ComponentDescriptor<any, any>> | undefined;
/**
 * Back reference to Component.
 */
export interface XTagElement<P, S> extends Element {
    /**
     * Back reference from Element to Component that is always present in DEBUG mode.
     */
    dxtag?: Component<P, S>;
    /**
     * Back reference from Element to Component. There are two references so we can check in DEBUG mode if we are missing
     * this backref in delegated event handlers.
     */
    xtag?: Component<P, S>;
}
/**
 * Component Descriptor.
 *
 * Each component should declare its properties and behavior in `ComponentDescriptor` object.
 *
 * Component descriptor has two parametric types: first parametric type `P` is a props type and second type `S` is a
 * state type.
 *
 * Component descriptor provides a `createComponent` and `createRootComponent` methods to create component instances.
 *
 * Example:
 *
 *     class State {
 *       xy: number;
 *
 *       constructor(props: number) {
 *         this.xy = props * props;
 *       }
 *     }
 *
 *     const MyComponent = new ComponentDescriptor<number, State>()
 *       .canvas()
 *       .createState((c, props) => new State(props))
 *       .update((c, props, state) => {
 *         const ctx = c.get2DContext();
 *         ctx.fillStyle = 'rgba(0, 0, 0, 1)';
 *         ctx.fillRect(props, props, state.xy, state.xy);
 *       });
 *
 *       const componentInstance = MyComponent.createRootComponent(10);
 *
 * @final
 */
export declare class ComponentDescriptor<P, S> {
    /**
     * Flags marked on component when component is instantiated. See `ComponentFlags` for details.
     */
    _markFlags: number;
    /**
     * Flags, see `ComponentDescriptorFlags` for details.
     */
    _flags: number;
    /**
     * Tag name of the root element or reference to a model.
     */
    _tag: string | VModel<any>;
    /**
     * New props received handler overrides default props received behavior and it should mark component as dirty if new
     * received props will cause change in component's representation.
     */
    _newPropsReceived: ((component: Component<P, S>, oldProps: P, newProps: P) => void) | null;
    /**
     * Lifecycle handler update.
     */
    _update: ((component: Component<P, S>, props: P, state: S) => void) | null;
    /**
     * Lifecycle handler createState should create a new state.
     */
    _createState: ((component: Component<P, S>, props: P) => S) | null;
    /**
     * Lifecycle handler init.
     */
    _init: ((component: Component<P, S>, props: P, state: S) => void) | null;
    /**
     * Lifecycle handler attached.
     */
    _attached: ((component: Component<P, S>, props: P, state: S) => void) | null;
    /**
     * Lifecycle handler detached.
     */
    _detached: ((component: Component<P, S>, props: P, state: S) => void) | null;
    /**
     * Lifecycle hdnaler disposed.
     */
    _disposed: ((component: Component<P, S>, props: P, state: S) => void) | null;
    /**
     * Pool of recycled components.
     */
    _recycledPool: Component<P, S>[] | null;
    /**
     * Maximum number of recycled components (recycled pool size).
     */
    _maxRecycled: number;
    /**
     * Name that is used in DEBUG mode.
     */
    name: string;
    constructor(name?: string);
    /**
     * Set tag name for the root element.
     *
     *     const MyComponent = new ComponentDescriptor()
     *       .tagName("table");
     */
    tagName(tagName: string): ComponentDescriptor<P, S>;
    /**
     * Use SVG Namespace to create a root element.
     *
     *     const MySvgComponent = new ComponentDescriptor()
     *       .svg()
     *       .tagName("circle");
     */
    svg(): ComponentDescriptor<P, S>;
    /**
     * Set VModel for the root element.
     *
     *     const MyComponent = new ComponentDescriptor()
     *       .vModel(new VModel("div").attrs({"id": "model"}))
     *       .update((c) => { c.vSync(c.createVRoot().children("content")); });
     */
    vModel(model: VModel<any>): ComponentDescriptor<P, S>;
    /**
     * Turn component into a canvas.
     *
     *     const MyCanvasComponent = new ComponentDescriptor()
     *       .canvas()
     *       .update((c) => {
     *         const ctx = c.get2DContext();
     *         ctx.fillStyle = "red";
     *         ctx.fillRect(0, 0, 10, 10);
     *       });
     */
    canvas(): ComponentDescriptor<P, S>;
    /**
     * Set lifecycle handler createState.
     *
     * Create state handler should return a new state. It is invoked immediately after component instantiation.
     *
     *     const MyComponent = new ComponentDescriptor<number, number>()
     *       .createState((c, props) => props * props)
     *       .update((c, props, state) => {
     *         c.vSync(c.createVRoot().children(state.toString()));
     *       });
     */
    createState(createState: (component: Component<P, S>, props: P) => S): ComponentDescriptor<P, S>;
    /**
     * Set lifecycle handler init.
     *
     * `element` and `props` properties will be initialized before init handler is invoked.
     *
     *     const MyComponent = new ComponentDescriptor()
     *       .tagName("button")
     *       .init((c) => {
     *         c.element.addEventListener("click", (e) => {
     *           e.preventDefault();
     *           e.stopPropagation();
     *           console.log("clicked");
     *         });
     *       })
     *       .update((c) => {
     *         c.vSync(c.createVRoot().children("click me"));
     *       });
     */
    init(init: (component: Component<P, S>, props: P, state: S) => void): ComponentDescriptor<P, S>;
    /**
     * Set newPropsReceived handler.
     *
     * New props received handler overrides default props received behavior and it should mark component as dirty if new
     * received props will cause change in component's representation.
     *
     *     const MyComponent = new ComponentDescriptor<{a: number}, void>()
     *       .newPropsReceived((c, oldProps, newProps) => {
     *         if (oldProps.a !== newProps.a) {
     *           c.markDirty();
     *         }
     *       })
     *       .update((c, props) => {
     *         c.vSync(c.createVRoot().children(c.props.toString()));
     *       });
     */
    newPropsReceived(newPropsReceived: (component: Component<P, S>, oldProps: P, newProps: P) => void): ComponentDescriptor<P, S>;
    /**
     * Set lifecycle handler update.
     *
     * Update handler overrides default update behavior.
     *
     *     const MyComponent = new ComponentDescriptor<{a: number}, void>()
     *       .update((c) => {
     *         c.vSync(c.createVRoot().children("content"));
     *       });
     */
    update(update: (component: Component<P, S>, props: P, state: S) => void): ComponentDescriptor<P, S>;
    /**
     * Set lifecycle handler attached.
     *
     * Attached handler will be invoked when component is attached to the document.
     *
     *     const onChange = new Invalidator();
     *
     *     const MyComponent = new ComponentDescriptor()
     *       .attached((c) => {
     *         c.subscribe(onChange);
     *       })
     *       .update((c) => {
     *         c.vSync(c.createVRoot().children("content"));
     *       });
     */
    attached(attached: (component: Component<P, S>, props: P, state: S) => void): ComponentDescriptor<P, S>;
    /**
     * Set lifecycle handler detached.
     *
     * Detached handler will be invoked when component is detached from the document.
     *
     *     const MyComponent = new ComponentDescriptor<any, {onResize: (e: Event) => void}>()
     *       .createState((c) => ({onResize: (e) => { console.log("window resized"); }})
     *       .attached((c, props, state) => {
     *         window.addEventListener("resize", state.onResize);
     *       })
     *       .detached((c, props, state) => {
     *         window.removeEventListener(state.onResize);
     *       })
     *       .update((c) => {
     *         c.vSync(c.createVRoot().children("content"));
     *       });
     */
    detached(detached: (component: Component<P, S>, props: P, state: S) => void): ComponentDescriptor<P, S>;
    /**
     * Set lifecycle handler disposed.
     *
     * Disposed handler will be invoked when component is disposed.
     *
     *     let allocatedComponents = 0;
     *
     *     const MyComponent = new ComponentDescriptor()
     *       .init((c) => {
     *         allocatedComponents++;
     *       })
     *       .disposed((c) => {
     *         allocatedComponents--;
     *       })
     *       .update((c) => {
     *         c.vSync(c.createVRoot().children("content"));
     *       });
     */
    disposed(disposed: (component: Component<P, S>, props: P, state: S) => void): ComponentDescriptor<P, S>;
    /**
     * Enable back reference from DOM element to component instance.
     *
     * Back reference will be assigned to `xtag` property.
     *
     *     const MyComponent = new ComponentDescriptor()
     *       .enableBackRef()
     *       .init((c) => {
     *         c.element.addEventListener("click", (e) => {
     *           console.log(getBackRef<Component>(e.target) === c);
     *         });
     *       })
     *       .update((c) => {
     *         c.vSync(c.createVRoot().children("content"));
     *       });
     */
    enableBackRef(): ComponentDescriptor<P, S>;
    /**
     * Enable component recycling.
     *
     * When component recycling is enabled, components will be instantiated from recycled pool.
     *
     *     const MyComponent = new ComponentDescriptor()
     *       .enableRecycling(100)
     *       .update((c) => {
     *         c.vSync(c.createVRoot().children("content"));
     *       });
     */
    enableRecycling(maxRecycled: number): ComponentDescriptor<P, S>;
    /**
     * Component has immutable props.
     */
    immutableProps(): ComponentDescriptor<P, S>;
    /**
     * Create a Virtual DOM node.
     *
     *     const MyComponent = new ComponentDescriptor<number, void>()
     *       .update((c, props) => {
     *         c.vSync(c.createVRoot().children(props.toString()));
     *       });
     *
     *     const vnode = MyComponent.createVNode(10);
     */
    createVNode(props?: P): VNode;
    /**
     * Create a Virtual DOM node with immutable props.
     *
     * Immutable props will be checked for identity before triggering an update, if props identity is the same, then
     * component won't be updated.
     *
     *     const MyComponent = new ComponentDescriptor<number, void>()
     *       .update((c, props) => {
     *         c.vSync(c.createVRoot().children(props.toString()));
     *       });
     *
     *     const vnode = MyComponent.createImmutableVNode(10);
     */
    createImmutableVNode(props?: P): VNode;
    /**
     * Creates a component instance without parent.
     *
     *     const MyComponent = new ComponentDescriptor<number, void>()
     *       .update((c, props) => {
     *         c.vSync(c.createVRoot().children(props.toString()));
     *       });
     *
     *     const component = MyComponent.createRootComponent(10);
     */
    createRootComponent(props?: P): Component<P, S>;
    /**
     * Create a Component.
     *
     *     const ChildComponent = new ComponentDescriptor()
     *       .update((c) => {
     *         c.vSync(c.createVRoot().children("child"));
     *       });
     *
     *     const ParentComponent = new ComponentDescriptor()
     *       .init((c) => {
     *         const child = ChildComponent.createComponent(c);
     *         c.element.appendChild(child.element);
     *         child.attached();
     *         child.update();
     *       });
     *
     *     const rootComponent = MyComponent.createRootComponent();
     */
    createComponent(parent: Component<any, any> | undefined, props?: P): Component<P, S>;
    /**
     * Mount component on top of existing html element.
     *
     *     const element = document.createElement("div");
     *     document.body.appendChild(element);
     *     element.innerHTML("<span>content</span>");
     *
     *     const MyComponent = new ComponentDescriptor()
     *       .update((c) => {
     *         c.vSync(c.createVRoot.children([createVElement("span").children("content")]));
     *       });
     *
     *     const component = MyComponent.mountComponent(element);
     *     component.update();
     */
    mountComponent(element: Element, parent?: Component<any, any>, props?: P): Component<P, S>;
    /**
     * Create event handler.
     */
    createEventHandler(handler: (event: Event, component: Component<P, S>, props: P, state: S) => void): (event: Event) => void;
    /**
     * Create delegated event handler.
     */
    createDelegatedEventHandler(selector: string | SelectorFn, componentSelector: string | SelectorFn | boolean, handler: (event: Event, component: Component<P, S>, props: P, state: S, matchingTarget: Element) => void): (event: Event) => void;
}
/**
 * Component.
 *
 * @final
 */
export declare class Component<P, S> {
    /**
     * Flags, see `ComponentFlags` for details.
     *
     * Lowest 24 bits are reserved for kivi flags, other bits can be used for user flags.
     */
    flags: number;
    /**
     * Timestamp when component were updated last time, using scheduler monotonically increasing clock.
     */
    mtime: number;
    /**
     * Component descriptor.
     */
    readonly descriptor: ComponentDescriptor<P, S>;
    /**
     * Reference to the root element.
     */
    readonly element: XTagElement<P, S>;
    /**
     * Depth in the components tree.
     *
     * Depth property is used by scheduler to determine its priority when updating components.
     */
    depth: number;
    /**
     * Component's props.
     */
    props: P | null;
    /**
     * Component's state.
     */
    state: S | null;
    /**
     * Root node can contain a virtual dom root if Component represents a DOM subtree, or Canvas context if Component is
     * a Canvas object.
     */
    _root: VNode | CanvasRenderingContext2D | null;
    _subscriptions: InvalidatorSubscription | null;
    _transientSubscriptions: InvalidatorSubscription | null;
    constructor(flags: number, descriptor: ComponentDescriptor<P, S>, element: Element, parent?: Component<any, any>, props?: P);
    /**
     * Get canvas 2d rendering context.
     */
    get2DContext(): CanvasRenderingContext2D;
    /**
     * Mark component as dirty and cancel all transient subscriptions.
     */
    markDirty(): void;
    /**
     * Creates a virtual dom root node.
     */
    createVRoot(): VNode;
    /**
     * **EXPERIMENTAL** Start interaction.
     *
     * When interaction is started, component becomes a high priority target for a scheduler, and scheduler goes into
     * throttled mode.
     */
    startInteraction(): void;
    /**
     * **EXPERIMENTAL** Finish interaction.
     *
     * Removes high priority flag from component and disables scheduler throttling.
     */
    finishInteraction(): void;
    /**
     * Adds component to a scheduler queue that will update component each animation frame.
     *
     * Component will be updated always, even when scheduler is in throttled mode.
     */
    startUpdateEachFrame(): void;
    /**
     * Remove component from a scheduler queue that updates component each animation frame.
     */
    stopUpdateEachFrame(): void;
    /**
     * Set new props and update component.
     *
     * Props are checked by their identity, unless it is disabled by component descriptor method
     * `disableCheckDataIdentity()`.
     */
    update(newProps?: P): void;
    /**
     * Sync internal representation using Virtual DOM API.
     *
     * If this method is called during mounting phase, then Virtual DOM will be mounted on top of the existing document
     * tree.
     */
    vSync(newRoot: VNode): void;
    /**
     * Invalidate component and schedule component to update on the next frame.
     *
     * Dirty flags parameter can be used to add hints that describe what has been changed.
     * This metho will automatically cancel all transient subscriptions if preserve transient subscriptions is false.
     */
    invalidate(dirtyFlags?: number, preserveTransientSubscriptions?: boolean): void;
    /**
     * Attach method should be invoked when component is attached to the document.
     */
    attach(): void;
    /**
     * Detach method should be invoked when component is detached from the document.
     */
    detach(): void;
    /**
     * Dispose method should be invoked when component is destroyed.
     */
    dispose(): void;
    /**
     * Subscribe to invalidator object.
     */
    subscribe(invalidator: Invalidator): InvalidatorSubscription;
    /**
     * Transiently subscribe to invalidator object.
     *
     * Each time component is invalidated, all transient subscriptions will be canceled.
     */
    transientSubscribe(invalidator: Invalidator): InvalidatorSubscription;
}
export declare function updateComponent(component: Component<any, any>, newProps?: any): void;
/**
 * Inject component into DOM.
 */
export declare function injectComponent<P, S>(descriptor: ComponentDescriptor<P, S>, container: Element, props?: P, sync?: boolean): Component<P, S>;
/**
 * Mount component on top of existing DOM.
 */
export declare function mountComponent<P, S>(descriptor: ComponentDescriptor<P, S>, element: Element, props?: P, sync?: boolean): Component<P, S>;
