import { Component } from "./component";
import { VNode } from "./vnode";
export declare type SchedulerTask = () => void;
/**
 * Frame tasks group contains tasks for updating components, read dom and write dom tasks, and tasks that should be
 * executed after all other tasks are finished.
 *
 * To get access to the frame tasks group, use: `currentFrame()` and `nextFrame()` scheduler methods.
 *
 *     scheduler.currentFrame().read(() => {
 *       console.log(element.clientWidth);
 *     });
 *
 * @final
 */
export declare class FrameTasksGroup {
    /**
     * See `FrameTasksGroupFlags` for details.
     */
    _flags: number;
    /**
     * Array of component arrays indexed by their depth.
     */
    _componentTasks: Array<Component<any, any>[] | null>;
    /**
     * Write DOM task queue.
     */
    _writeTasks: SchedulerTask[] | null;
    /**
     * Read DOM task queue.
     */
    _readTasks: SchedulerTask[] | null;
    /**
     * Tasks that should be executed when all other tasks are finished.
     */
    _afterTasks: SchedulerTask[] | null;
    /**
     * Element that should be focused.
     */
    _focus: Element | VNode | null;
    constructor();
    /**
     * Add Component to the components queue.
     */
    updateComponent(component: Component<any, any>): void;
    /**
     * Add new task to the write DOM task queue.
     */
    write(callback: SchedulerTask): void;
    /**
     * Add new task to the read DOM task queue.
     */
    read(callback: SchedulerTask): void;
    /**
     * Add new task to the task queue that will execute tasks when all DOM tasks are finished.
     */
    after(callback: SchedulerTask): void;
    /**
     * Set a focus on an element when all DOM tasks are finished.
     */
    focus(node: Element | VNode): void;
    /**
     * Place a lock on adding new read and write task.
     *
     * Works in DEBUG mode only.
     */
    _rwLock(): void;
    /**
     * Remove a lock from adding new read and write tasks.
     *
     * Works in DEBUG mode only.
     */
    _rwUnlock(): void;
}
/**
 * Returns current monotonically increasing clock.
 */
export declare function clock(): number;
/**
 * Add task to the microtask queue.
 */
export declare function scheduleMicrotask(task: () => void): void;
/**
 * Add task to the macrotask queue.
 */
export declare function scheduleMacrotask(task: () => void): void;
/**
 * Get task list for the current frame.
 */
export declare function currentFrame(): FrameTasksGroup;
/**
 * Get task list for the next frame.
 */
export declare function nextFrame(): FrameTasksGroup;
/**
 * **EXPERIMENTAL** Enable throttling mode.
 *
 * In throttling mode, all updates for low priority components will be throttled, all updates will be performed
 * incrementally each frame.
 *
 * Each time throttling is enabled, it increases internal counter that tracks how many times it is enabled, and when
 * counters goes to zero, throttling mode will be disabled.
 */
export declare function enableThrottling(): void;
/**
 * **EXPERIMENTAL** Disable throttling mode.
 *
 * Throttling mode will be disabled when number of dependencies that enabled throttling mode goes to zero.
 */
export declare function disableThrottling(): void;
/**
 * Get remaining time available for tasks in the current frame.
 */
export declare function frameTimeRemaining(): number;
/**
 * Add component to the list of components that should be updated each frame.
 */
export declare function startUpdateComponentEachFrame(component: Component<any, any>): void;
/**
 * Returns true if UI state is in mounting phase.
 */
export declare function isMounting(): boolean;
/**
 * Returns true if tasks should be throttled.
 */
export declare function isThrottled(): boolean;
/**
 * Start UI mounting phase.
 */
export declare function startMounting(): void;
/**
 * Finish UI mounting phase.
 */
export declare function finishMounting(): void;
