import { VNode } from "./vnode";
/**
 * Model for DOM Elements.
 *
 * Models are used as an advanced optimization technique. When creating virtual nodes, or declaring root node for a
 * component, it is possible to link them to a model instead of simple HTML tagName. Model will contain all static
 * properties for the HTML Element, so there is no need to declare them each time virtual node is created. It also
 * reduces diff overhead, because there is no need to diff static model properties.
 *
 * Creating a virtual dom node from VModel:
 *
 *     const model = new VModel("div").attrs({"id": "model"});
 *     // Element node
 *     const node = mode.createVNode();
 *     // Component's root node
 *     const root = mode.createVRoot();
 *
 * Creating a DOM node from VModel:
 *
 *     const model = new VModel("div").attrs({"id": "model"});
 *     const div = model.createElement();
 *
 * @final
 */
export declare class VModel<D> {
    /**
     * Flags that should be marked on VNode and ComponentDescriptor when they are associated with VModel.
     *
     * See `SharedFlags` for details.
     */
    _markFlags: number;
    /**
     * Flags, see `VModelFlags` for details.
     */
    _flags: number;
    /**
     * Tag name of the element.
     */
    _tagName: string;
    /**
     * Properties.
     *
     * All properties are assigned to DOM nodes directly:
     *
     *     e: HTMLElement;
     *     e.propertyName = propertyValue;
     *
     * When virtual node is mounted on top of existing HTML, all properties from model will be assigned during mounting
     * phase.
     */
    _props: {
        [key: string]: any;
    } | null;
    /**
     * Attributes.
     *
     * All attributes are assigned to DOM nodes with `setAttribute` method:
     *
     *     e: HTMLElement;
     *     e.setAttribute(key, value);
     *
     * If attribute is prefixed with "xlink:", or "xml:" namespace, it will assign attributes with `setAttributeNS`
     * method and use appropriate namespaces.
     */
    _attrs: {
        [key: string]: any;
    } | null;
    /**
     * Style in css string format.
     *
     * Style is assigned to DOM nodes with `style.cssText` property, if virtual node represents an element from svg
     * namespace, style will be assigned with `setAttribute("style", "cssText")` method.
     */
    _style: string | null;
    /**
     * Class name.
     *
     * Class name is assigned to DOM nodes with `className` property, if virtual node represents an element from svg
     * namespace, class name will be assigned with `setAttribute("class", "className")` method.
     */
    _className: string | null;
    /**
     * Update handler is used to override default reconciliation algorithm.
     */
    private _updateHandler;
    /**
     * Reference to an element that will be cloned when DOM node cloning is enabled.
     */
    private _ref;
    constructor(tagName: string);
    /**
     * Use svg namespace for the DOM element.
     */
    svg(): VModel<D>;
    /**
     * Set properties.
     *
     * All properties are assigned to DOM nodes directly:
     *
     *     e: HTMLElement;
     *     e.propertyName = propertyValue;
     *
     * When virtual node is mounted on top of existing HTML, all properties from model will be assigned during mounting
     * phase.
     */
    props(props: {
        [key: string]: any;
    }): VModel<D>;
    /**
     * Set attributes.
     *
     * All attributes are assigned to DOM nodes with `setAttribute` method:
     *
     *     e: HTMLElement;
     *     e.setAttribute(key, value);
     *
     * If attribute is prefixed with "xlink:", or "xml:" namespace, it will assign attributes with `setAttributeNS`
     * method and use appropriate namespaces.
     */
    attrs(attrs: {
        [key: string]: any;
    }): VModel<D>;
    /**
     * Set style in css string format.
     *
     * Style is assigned to DOM nodes with `style.cssText` property, if virtual node represents an element from svg
     * namespace, style will be assigned with `setAttribute("style", "cssText")` method.
     */
    style(style: string): VModel<D>;
    /**
     * Set className.
     *
     * Class name is assigned to DOM nodes with `className` property, if virtual node represents an element from svg
     * namespace, class name will be assigned with `setAttribute("class", "className")` method.
     */
    className(classes: string): VModel<D>;
    /**
     * Enable node cloning.
     *
     * Instead of creating DOM nodes, model will clone nodes from a base node with `Node.cloneNode(false)` method.
     */
    enableCloning(): VModel<D>;
    /**
     * Set update handler.
     *
     * Update handler is used to override default reconciliation algorithm.
     */
    updateHandler(handler: VModelUpdateHandler<D>): VModel<D>;
    /**
     * Create a Virtual DOM Node.
     */
    createVNode(data?: D): VNode;
    /**
     * Create a Virtual DOM Node for component's root.
     */
    createVRoot(data?: D): VNode;
    /**
     * Create a DOM Element.
     */
    createElement(): Element;
    /**
     * Update DOM Node with an update handler.
     */
    update(element: Element, oldProps: D | undefined, newProps: D): void;
}
/**
 * VModel update handler is used to override default reconciliation algorithm.
 *
 * When `oldProps` is `undefined`, it means that element is created.
 */
export declare type VModelUpdateHandler<D> = (element: Element, oldProps: D | undefined, newProps: D) => void;
