import { VNodeFlags } from "./misc";
import { Component, ComponentDescriptor } from "./component";
import { VModel } from "./vmodel";
/**
 * Virtual DOM Node.
 *
 * VNode object is the core object in kivi Virtual DOM, it can represent any node type.
 *
 * The easiest way to create VNode instances is to use factory functions:
 *
 *     // HTMLElement
 *     const htmlElement = createVElement("div");
 *     // SVGElement
 *     const svgElement = createVSvgElement("a");
 *     // Text node
 *     const textNode = createVText("text content");
 *     // Component's root node
 *     const root = createVRoot();
 *
 * VNode instances can't be reused to represent multiple DOM nodes.
 *
 * @final
 */
export declare class VNode {
    /**
     * Flags, see `VNodeFlags` for details.
     */
    _flags: VNodeFlags;
    /**
     * Tag name of the element, or reference to VModel if virtual node represents an element, or ComponentDescriptor
     * if it represents a component.
     */
    _tag: string | VModel<any> | ComponentDescriptor<any, any> | null;
    /**
     * Children reconciliation algorithm is using key property to find the same node in the previous children array. Key
     * should be unique among its siblings.
     */
    _key: any;
    /**
     * Properties.
     *
     * When virtual node represents an element, props property is used to set properties directly on DOM node:
     *
     *     e: HTMLElement;
     *     e.propertyName = propertyValue;
     *
     * When virtual node is mounted on top of existing HTML, all properties will be assigned during mounting phase.
     */
    _props: any;
    /**
     * Attributes.
     *
     * All attributes are assigned to DOM nodes with `setAttribute` method:
     *
     *     e: HTMLElement;
     *     e.setAttribute(key, value);
     *
     * If attribute is prefixed with "xlink:", or "xml:" namespace, it will assign attributes with `setAttributeNS`
     * method and use appropriate namespaces.
     */
    _attrs: {
        [key: string]: any;
    } | null;
    /**
     * Style in css string format.
     *
     * Style is assigned to DOM nodes with `style.cssText` property, if virtual node represents an element from svg
     * namespace, style will be assigned with `setAttribute("style", "cssText")` method.
     */
    _style: string | null;
    /**
     * Class name.
     *
     * Class name is assigned to DOM nodes with `className` property, if virtual node represents an element from svg
     * namespace, class name will be assigned with `setAttribute("class", "className")` method.
     */
    _className: string | null;
    /**
     * Children property can contain flat array of children virtual nodes, or text if it contains a single text node
     * child. If virtual node represents an input field, children property will contain input value.
     */
    _children: VNode[] | string | boolean | null;
    /**
     * Reference to HTML Node. It will be available after virtual node is created or synced. Each time VNode is synced,
     * reference to the HTML Node is transferred from old virtual node to the new one.
     */
    ref: Node | null;
    /**
     * Reference to a Component. If virtual node is a Component, then cref will be available after virtual node is
     * created or synced. Each time virtual node is synced, reference to a Component is transferred from old virtual
     * node to the new one.
     */
    cref: Component<any, any> | null;
    /**
     * Debug properties are used because VNode properties are frozen.
     *
     * See `VNodeDebugFlags` for details.
     */
    _debugProperties: {
        flags: number;
    };
    constructor(flags: number, tag: string | VModel<any> | ComponentDescriptor<any, any> | null, props: any);
    /**
     * Set key.
     *
     * Children reconciliation algorithm is using key property to find the same node in the previous children array. Key
     * should be unique among its siblings.
     *
     * This method is available in all virtual node types.
     */
    key(key: any): VNode;
    /**
     * Set properties with static shape.
     *
     * Each time virtual node representing the same DOM node is created, it should have properties with exactly the same
     * shape. Values can be different, but all keys should be the same. For example:
     *
     *     const a = createVElement("div").props({"id": "Main", "title": "Title"});
     *     const b = createVElement("div").props({"id": "Main", "title": "New Title"});
     *
     * Props property is used to set properties directly on DOM node:
     *
     *     e: HTMLElement;
     *     e.propertyName = propertyValue;
     *
     * When virtual node is mounted on top of existing HTML, all properties will be assigned during mounting phase.
     *
     * If virtual node is using `VModel` instance with custom update handler, update data should be assigned with `data`
     * method.
     *
     * This method is available on element and component's root virtual node types.
     */
    props(props: {
        [key: string]: any;
    }): VNode;
    /**
     * Set attributes with static shape.
     *
     * Each time virtual node representing the same DOM node is created, it should have attributes with exactly the same
     * shape. Values can be different, but all keys should be the same. For example:
     *
     *     const a = createVElement("div").attrs({"id": "Main", "title": "Title"});
     *     const b = createVElement("div").attrs({"id": "Main", "title": "New Title"});
     *
     * All attributes are assigned to DOM nodes with `setAttribute` method:
     *
     *     e: HTMLElement;
     *     e.setAttribute(key, value);
     *
     * If attribute is prefixed with "xlink:", or "xml:" namespace, it will assign attributes with `setAttributeNS`
     * method and use appropriate namespaces.
     *
     * If virtual node is using `VModel` instance with custom update handler, update data should be assigned with `data`
     * method.
     *
     * This method is available on element and component's root virtual node types.
     */
    attrs(attrs: {
        [key: string]: any;
    }): VNode;
    /**
     * Set attributes with dynamic shape.
     *
     * Attributes can have different set of keys, when reconciliation algorithm updates attribute values, it will remove
     * attributes with `removeAttribute` method for all missing keys.
     *
     *     const a = createVElement("div").props({"id": "Main", "title": "Title"});
     *     const b = createVElement("div").props({"id": "Main"});
     *
     * All attributes are assigned to DOM nodes with `setAttribute` method:
     *
     *     e: HTMLElement;
     *     e.setAttribute(key, value);
     *
     * If attribute is prefixed with "xlink:", or "xml:" namespace, it will assign attributes with `setAttributeNS`
     * method and use appropriate namespaces.
     *
     * If virtual node is using `VModel` instance with custom update handler, update data should be assigned with `data`
     * method.
     *
     * This method is available on element and component's root virtual node types.
     */
    dynamicShapeAttrs(attrs?: {
        [key: string]: any;
    }): VNode;
    /**
     * Set update data for VModel custom update handler.
     *
     * This method is available on element and component's root virtual node types.
     */
    data(data: any): VNode;
    /**
     * Set style in css string format.
     *
     * Style is assigned to DOM nodes with `style.cssText` property, if virtual node represents an element from svg
     * namespace, style will be assigned with `setAttribute("style", "cssText")` method.
     *
     * This method is available on element and component's root virtual node types.
     */
    style(style: string): VNode;
    /**
     * Set className.
     *
     * Class name is assigned to DOM nodes with `className` property, if virtual node represents an element from svg
     * namespace, class name will be assigned with `setAttribute("class", "className")` method.
     *
     * This method is available on element, component and component's root virtual node types.
     */
    className(className: string): VNode;
    /**
     * Set children. Children parameter should be either a flat array of virtual nodes, or text if node contains a single
     * text node.
     *
     * This method is available on element and component's root virtual node types.
     */
    children(children: VNode[] | string): VNode;
    /**
     * Set children as an innerHTML string. It is potentially vulnerable to XSS attacks.
     *
     * This method is available on element and component's root virtual node types.
     */
    unsafeHTML(html: string): VNode;
    /**
     * Set children and enable track by key reconciliation algorithm. Children parameter should be a flat array of
     * virtual nodes with assigned key properties.
     *
     * This method is available on element and component's root virtual node types.
     */
    trackByKeyChildren(children: VNode[]): VNode;
    /**
     * Set text value for Input Elements.
     *
     * This method is available on text input element node type.
     */
    value(value: string): VNode;
    /**
     * Set checked value for Input Elements.
     *
     * This method is available on checked input element node type.
     */
    checked(value: boolean): VNode;
    /**
     * Prevents component node from syncing.
     *
     * This method is available on component node type.
     */
    bindOnce(): VNode;
    /**
     * Keep alive node when removing from the document. Keep alive nodes will be detached instead of disposed when they
     * are removed from virtual dom tree.
     *
     * Keep alive nodes should be manually disposed by owning component.
     */
    keepAlive(component: Component<any, any>): VNode;
    /**
     * Disable children shape errors in DEBUG mode.
     */
    disableChildrenShapeError(): VNode;
    getComponentRef<P, S>(): Component<P, S>;
}
/**
 * Recursively attach all nodes.
 */
export declare function vNodeAttach(vnode: VNode): void;
/**
 * This method should be invoked when node is attached, we don't use recursive implementation because we appending
 * nodes to the document as soon as they created and children nodes aren"t attached at this time.
 */
export declare function vNodeAttached(vnode: VNode): void;
/**
 * Recursively detach all nodes.
 */
export declare function vNodeDetach(vnode: VNode): void;
/**
 * Recursively dispose all nodes.
 */
export declare function vNodeDispose(vnode: VNode): void;
/**
 * Instantiate a DOM Node or Component from the Virtual DOM Node.
 *
 * This method doesn't set any attributes, or create children, to render internal representation of the virtual node,
 * use `vNodeRender` method.
 */
export declare function vNodeInstantiate(vnode: VNode, owner: Component<any, any> | undefined): void;
/**
 * Render internal representation of the Virtual DOM Node.
 */
export declare function vNodeRender(vnode: VNode, owner: Component<any, any> | undefined): void;
/**
 * Mount VNode on top of existing html document.
 */
export declare function vNodeMount(vnode: VNode, node: Node, owner: Component<any, any> | undefined): void;
/**
 * Sync two VNodes
 *
 * When node `a` is synced with node `b`, `a` node should be considered as destroyed, and any access to it after sync
 * is an undefined behavior.
 */
export declare function syncVNodes(a: VNode, b: VNode, owner?: Component<any, any>): void;
/**
 * Create a VNode representing a [Text] node.
 */
export declare function createVText(content: string): VNode;
/**
 * Create a VNode representing an [Element] node.
 */
export declare function createVElement(tagName: string): VNode;
/**
 * Create a VNode representing a [SVGElement] node.
 */
export declare function createVSvgElement(tagName: string): VNode;
/**
 * Create a VNode representing a Component root node.
 */
export declare function createVRoot(): VNode;
