(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (factory((global.kivi = global.kivi || {})));
}(this, function (exports) { 'use strict';

    /**
     * Print error to the console and throw local exception, so that we can break
     * on caught errors.
     */
    function printError(message) {
        console.error(message);
        try {
            throw new Error(message);
        }
        catch (_) {
        }
    }

    const SvgNamespace = "http://www.w3.org/2000/svg";
    const XlinkNamespace = "http://www.w3.org/1999/xlink";
    const XmlNamespace = "http://www.w3.org/XML/1998/namespace";
    /**
     * Set HTML attribute.
     */
    function setAttr(node, key, value) {
        if (key.charCodeAt(0) !== 120) {
            node.setAttribute(key, value);
        }
        else {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if (key.length <= 3) {
                    throw new Error(`Failed to set attr: invalid attribute "${key}", attributes starting with letter "x" should` +
                        ` have length 4 or more.`);
                }
            }
            if (key.charCodeAt(1) === 109 && key.charCodeAt(2) === 108) {
                if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                    if (key.charCodeAt(3) !== 58) {
                        throw new Error(`Failed to set attr: invalid attribute "${key}", attributes with "xml" prefix should be in` +
                            ` the form "xml:attr".`);
                    }
                }
                node.setAttributeNS(XmlNamespace, key, value);
            }
            else if (key.charCodeAt(1) === 108 && key.charCodeAt(2) === 105) {
                if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                    if (key.charCodeAt(3) !== 110 || key.charCodeAt(4) !== 107 || key.charCodeAt(5) !== 58) {
                        throw new Error(`Failed to set attr: invalid attribute "${key}", attributes with "xli" prefix should be in` +
                            ` the form "xlink:attr".`);
                    }
                }
                node.setAttributeNS(XlinkNamespace, key, value);
            }
            else {
                node.setAttribute(key, value);
            }
        }
    }
    /**
     * This function doesn't do anything, it just returns the same `tagName` value.
     *
     * It is used so that when building application in production mode, we can replace all occurences of `getTagName`
     * function and minify all tag names.
     */
    function getTagName(tagName) {
        return tagName;
    }
    /**
     * This function doesn't do anything, it just returns the same `className` value.
     *
     * It is used so that when building application in production mode, we can replace all occurences of `getClassName`
     * function and minify all class names.
     */
    function getClassName(className) {
        return className;
    }
    function matchesWithAncestors(element, selector, sentinel = null) {
        let e = element;
        do {
            if (typeof selector === "string") {
                if (e.matches(selector)) {
                    return e;
                }
            }
            else if (selector(e)) {
                return e;
            }
            e = e.parentElement;
        } while (e !== sentinel);
        return null;
    }

    /**
     * Virtual DOM Node.
     *
     * VNode object is the core object in kivi Virtual DOM, it can represent any node type.
     *
     * The easiest way to create VNode instances is to use factory functions:
     *
     *     // HTMLElement
     *     const htmlElement = createVElement("div");
     *     // SVGElement
     *     const svgElement = createVSvgElement("a");
     *     // Text node
     *     const textNode = createVText("text content");
     *     // Component's root node
     *     const root = createVRoot();
     *
     * VNode instances can't be reused to represent multiple DOM nodes.
     *
     * @final
     */
    class VNode {
        constructor(flags, tag, props) {
            this._flags = flags;
            this._tag = tag;
            this._key = null;
            this._props = props;
            this._attrs = null;
            this._style = null;
            this._className = null;
            this._children = null;
            this.ref = null;
            this.cref = null;
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                this._debugProperties = {
                    flags: 0,
                };
            }
        }
        /**
         * Set key.
         *
         * Children reconciliation algorithm is using key property to find the same node in the previous children array. Key
         * should be unique among its siblings.
         *
         * This method is available in all virtual node types.
         */
        key(key) {
            this._key = key;
            return this;
        }
        /**
         * Set properties with static shape.
         *
         * Each time virtual node representing the same DOM node is created, it should have properties with exactly the same
         * shape. Values can be different, but all keys should be the same. For example:
         *
         *     const a = createVElement("div").props({"id": "Main", "title": "Title"});
         *     const b = createVElement("div").props({"id": "Main", "title": "New Title"});
         *
         * Props property is used to set properties directly on DOM node:
         *
         *     e: HTMLElement;
         *     e.propertyName = propertyValue;
         *
         * When virtual node is mounted on top of existing HTML, all properties will be assigned during mounting phase.
         *
         * If virtual node is using `VModel` instance with custom update handler, update data should be assigned with `data`
         * method.
         *
         * This method is available on element and component's root virtual node types.
         */
        props(props) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & (2 /* Element */ | 8 /* Root */)) === 0) {
                    throw new Error("Failed to set props on VNode: props method should be called on element or component" +
                        " root nodes only.");
                }
                if ((this._flags & 131072 /* VModel */) !== 0) {
                    if ((this._flags & 262144 /* VModelUpdateHandler */) !== 0) {
                        throw new Error("Failed to set props on VNode: VNode is using VModel with custom update handler.");
                    }
                    const model = this._tag;
                    if (model._props !== null) {
                        const keys = Object.keys(props);
                        for (let i = 0; i < keys.length; i++) {
                            if (model._props.hasOwnProperty(keys[i])) {
                                throw new Error(`Failed to set props on VNode: VNode is using VModel that uses the same` +
                                    ` property "${keys[i]}".`);
                            }
                        }
                    }
                }
            }
            this._props = props;
            return this;
        }
        /**
         * Set attributes with static shape.
         *
         * Each time virtual node representing the same DOM node is created, it should have attributes with exactly the same
         * shape. Values can be different, but all keys should be the same. For example:
         *
         *     const a = createVElement("div").attrs({"id": "Main", "title": "Title"});
         *     const b = createVElement("div").attrs({"id": "Main", "title": "New Title"});
         *
         * All attributes are assigned to DOM nodes with `setAttribute` method:
         *
         *     e: HTMLElement;
         *     e.setAttribute(key, value);
         *
         * If attribute is prefixed with "xlink:", or "xml:" namespace, it will assign attributes with `setAttributeNS`
         * method and use appropriate namespaces.
         *
         * If virtual node is using `VModel` instance with custom update handler, update data should be assigned with `data`
         * method.
         *
         * This method is available on element and component's root virtual node types.
         */
        attrs(attrs) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & (2 /* Element */ | 8 /* Root */)) === 0) {
                    throw new Error("Failed to set attrs on VNode: attrs method should be called on element or component" +
                        " root nodes only.");
                }
                if ((this._flags & 131072 /* VModel */) !== 0) {
                    if ((this._flags & 262144 /* VModelUpdateHandler */) !== 0) {
                        throw new Error("Failed to set attrs on VNode: VNode is using VModel with custom update handler.");
                    }
                    const model = this._tag;
                    if (model._attrs !== null) {
                        const keys = Object.keys(attrs);
                        for (let i = 0; i < keys.length; i++) {
                            if (model._attrs.hasOwnProperty(keys[i])) {
                                throw new Error(`Failed to set attrs on VNode: VNode is using VModel that uses the same` +
                                    ` attribute "${keys[i]}".`);
                            }
                        }
                    }
                }
            }
            this._attrs = attrs;
            return this;
        }
        /**
         * Set attributes with dynamic shape.
         *
         * Attributes can have different set of keys, when reconciliation algorithm updates attribute values, it will remove
         * attributes with `removeAttribute` method for all missing keys.
         *
         *     const a = createVElement("div").props({"id": "Main", "title": "Title"});
         *     const b = createVElement("div").props({"id": "Main"});
         *
         * All attributes are assigned to DOM nodes with `setAttribute` method:
         *
         *     e: HTMLElement;
         *     e.setAttribute(key, value);
         *
         * If attribute is prefixed with "xlink:", or "xml:" namespace, it will assign attributes with `setAttributeNS`
         * method and use appropriate namespaces.
         *
         * If virtual node is using `VModel` instance with custom update handler, update data should be assigned with `data`
         * method.
         *
         * This method is available on element and component's root virtual node types.
         */
        dynamicShapeAttrs(attrs) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & (2 /* Element */ | 8 /* Root */)) === 0) {
                    throw new Error("Failed to set attrs on VNode: attrs method should be called on element or component" +
                        " root nodes only.");
                }
                if ((this._flags & 131072 /* VModel */) !== 0) {
                    if ((this._flags & 262144 /* VModelUpdateHandler */) !== 0) {
                        throw new Error("Failed to set attrs on VNode: VNode is using VModel with custom update handler.");
                    }
                    const model = this._tag;
                    if (model._attrs !== null) {
                        const keys = Object.keys(attrs);
                        for (let i = 0; i < keys.length; i++) {
                            if (model._attrs.hasOwnProperty(keys[i])) {
                                throw new Error(`Failed to set attrs on VNode: VNode is using VModel that uses the same` +
                                    ` attribute "${keys[i]}".`);
                            }
                        }
                    }
                }
            }
            this._flags |= 32 /* DynamicShapeAttrs */;
            this._attrs = attrs === undefined ? null : attrs;
            return this;
        }
        /**
         * Set update data for VModel custom update handler.
         *
         * This method is available on element and component's root virtual node types.
         */
        data(data) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & (2 /* Element */ | 8 /* Root */ | 131072 /* VModel */)) === 0) {
                    throw new Error("Failed to set data on VNode: data method should be called on element or component" +
                        " root nodes represented by VModel only.");
                }
                if ((this._flags & 262144 /* VModelUpdateHandler */) === 0) {
                    throw new Error("Failed to set data on VNode: VNode should be using VModel with custom update handler.");
                }
            }
            this._props = data;
            return this;
        }
        /**
         * Set style in css string format.
         *
         * Style is assigned to DOM nodes with `style.cssText` property, if virtual node represents an element from svg
         * namespace, style will be assigned with `setAttribute("style", "cssText")` method.
         *
         * This method is available on element and component's root virtual node types.
         */
        style(style) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & (2 /* Element */ | 8 /* Root */)) === 0) {
                    throw new Error("Failed to set style on VNode: style method should be called on element or component" +
                        " root nodes only.");
                }
                if ((this._flags & 131072 /* VModel */) !== 0) {
                    if ((this._flags & 262144 /* VModelUpdateHandler */) !== 0) {
                        throw new Error("Failed to set style on VNode: VNode is using VModel with custom update handler.");
                    }
                    if ((this._tag._style) !== null) {
                        throw new Error("Failed to set style on VNode: VNode is using VModel with assigned style.");
                    }
                }
            }
            this._style = style;
            return this;
        }
        /**
         * Set className.
         *
         * Class name is assigned to DOM nodes with `className` property, if virtual node represents an element from svg
         * namespace, class name will be assigned with `setAttribute("class", "className")` method.
         *
         * This method is available on element, component and component's root virtual node types.
         */
        className(className) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & (2 /* Element */ | 4 /* Component */ | 8 /* Root */)) === 0) {
                    throw new Error("Failed to set classes on VNode: classes method should be called on element or component" +
                        " root nodes only.");
                }
                if ((this._flags & 131072 /* VModel */) !== 0) {
                    if ((this._flags & 262144 /* VModelUpdateHandler */) !== 0) {
                        throw new Error("Failed to set style on VNode: vnode is using vmodel with custom update handler.");
                    }
                    if ((this._tag._className) !== null) {
                        throw new Error("Failed to set style on VNode: vnode is using vmodel with assigned className.");
                    }
                }
            }
            this._className = className;
            return this;
        }
        /**
         * Set children. Children parameter should be either a flat array of virtual nodes, or text if node contains a single
         * text node.
         *
         * This method is available on element and component's root virtual node types.
         */
        children(children) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & (2 /* Element */ | 8 /* Root */)) === 0) {
                    throw new Error("Failed to set children on VNode: children method should be called on element or component" +
                        " root nodes only.");
                }
                if ((this._flags & 192 /* InputElement */) !== 0) {
                    throw new Error("Failed to set children on VNode: input elements can't have children.");
                }
            }
            this._children = children;
            return this;
        }
        /**
         * Set children as an innerHTML string. It is potentially vulnerable to XSS attacks.
         *
         * This method is available on element and component's root virtual node types.
         */
        unsafeHTML(html) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & (2 /* Element */ | 8 /* Root */)) === 0) {
                    throw new Error("Failed to set unsafeHTML on VNode: unsafeHTML method should be called on element or" +
                        " component root nodes only.");
                }
                if ((this._flags & 192 /* InputElement */) !== 0) {
                    throw new Error("Failed to set unsafeHTML on VNode: input elements can't have children.");
                }
            }
            this._flags |= 2048 /* UnsafeHTML */;
            this._children = html;
            return this;
        }
        /**
         * Set children and enable track by key reconciliation algorithm. Children parameter should be a flat array of
         * virtual nodes with assigned key properties.
         *
         * This method is available on element and component's root virtual node types.
         */
        trackByKeyChildren(children) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & (2 /* Element */ | 8 /* Root */)) === 0) {
                    throw new Error("Failed to set children on VNode: children method should be called on element or component" +
                        " root nodes only.");
                }
                if ((this._flags & 192 /* InputElement */) !== 0) {
                    throw new Error("Failed to set children on VNode: input elements can't have children.");
                }
                if (children !== null) {
                    for (let i = 0; i < children.length; i++) {
                        if (children[i]._key === null) {
                            throw new Error("Failed to set children on VNode: trackByKeyChildren method expects all children to" +
                                " have a key.");
                        }
                    }
                }
            }
            this._flags |= 16 /* TrackByKeyChildren */;
            this._children = children;
            return this;
        }
        /**
         * Set text value for Input Elements.
         *
         * This method is available on text input element node type.
         */
        value(value) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if (this._children !== null) {
                    throw new Error("Failed to set value on VNode: VNode is already have either children or checked property.");
                }
            }
            this._flags |= 64 /* TextInputElement */;
            this._children = value;
            return this;
        }
        /**
         * Set checked value for Input Elements.
         *
         * This method is available on checked input element node type.
         */
        checked(value) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if (this._children !== null) {
                    throw new Error("Failed to set checked on VNode: VNode is already have either children or value property.");
                }
            }
            this._flags |= 128 /* CheckedInputElement */;
            this._children = value;
            return this;
        }
        /**
         * Prevents component node from syncing.
         *
         * This method is available on component node type.
         */
        bindOnce() {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & 4 /* Component */) === 0) {
                    throw new Error("Failed to set bindOnce on VNode: bindOnce method should be called on component nodes only.");
                }
            }
            this._flags |= 512 /* BindOnce */;
            return this;
        }
        /**
         * Keep alive node when removing from the document. Keep alive nodes will be detached instead of disposed when they
         * are removed from virtual dom tree.
         *
         * Keep alive nodes should be manually disposed by owning component.
         */
        keepAlive(component) {
            this._flags |= 256 /* KeepAlive */;
            this.ref = component.element;
            this.cref = component;
            return this;
        }
        /**
         * Disable children shape errors in DEBUG mode.
         */
        disableChildrenShapeError() {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & (2 /* Element */ | 8 /* Root */)) === 0) {
                    throw new Error("Failed to disable children shape error on VNode: disableChildrenShapeError method should" +
                        " be called on element or component root nodes only.");
                }
                this._debugProperties.flags |= 32 /* DisabledChildrenShapeError */;
            }
            return this;
        }
        getComponentRef() {
            return this.cref;
        }
    }
    /**
     * Recursively attach all nodes.
     */
    function vNodeAttach(vnode) {
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if ((vnode._debugProperties.flags & 4 /* Attached */) !== 0) {
                throw new Error("Failed to attach VNode: VNode is already attached.");
            }
            vnode._debugProperties.flags |= 4 /* Attached */;
            vnode._debugProperties.flags &= ~8 /* Detached */;
        }
        if ((vnode._flags & 4 /* Component */) === 0) {
            const children = vnode._children;
            if (children !== null && typeof children !== "string") {
                for (let i = 0; i < children.length; i++) {
                    vNodeAttach(children[i]);
                }
            }
        }
        else {
            vnode.cref.attach();
        }
    }
    /**
     * This method should be invoked when node is attached, we don't use recursive implementation because we appending
     * nodes to the document as soon as they created and children nodes aren"t attached at this time.
     */
    function vNodeAttached(vnode) {
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if ((vnode._debugProperties.flags & 4 /* Attached */) !== 0) {
                throw new Error("Failed to attach VNode: VNode is already attached.");
            }
            vnode._debugProperties.flags |= 4 /* Attached */;
            vnode._debugProperties.flags &= ~8 /* Detached */;
        }
        if ((vnode._flags & 4 /* Component */) !== 0) {
            vnode.cref.attach();
        }
    }
    /**
     * Recursively detach all nodes.
     */
    function vNodeDetach(vnode) {
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if ((vnode._debugProperties.flags & 8 /* Detached */) !== 0) {
                throw new Error("Failed to detach VNode: VNode is already detached.");
            }
            vnode._debugProperties.flags |= 8 /* Detached */;
            vnode._debugProperties.flags &= ~4 /* Attached */;
        }
        if ((vnode._flags & 4 /* Component */) === 0) {
            const children = vnode._children;
            if (children !== null && typeof children !== "string") {
                for (let i = 0; i < children.length; i++) {
                    vNodeDetach(children[i]);
                }
            }
        }
        else {
            vnode.cref.detach();
        }
    }
    /**
     * Recursively dispose all nodes.
     */
    function vNodeDispose(vnode) {
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if ((vnode._debugProperties.flags & 16 /* Disposed */) !== 0) {
                throw new Error("Failed to dispose VNode: VNode is already disposed.");
            }
            if ((vnode._debugProperties.flags & (1 /* Rendered */ | 2 /* Mounted */)) === 0) {
                throw new Error("Failed to dispose VNode: VNode should be rendered or mounted before disposing.");
            }
            vnode._debugProperties.flags |= 16 /* Disposed */;
        }
        if ((vnode._flags & 256 /* KeepAlive */) === 0) {
            if ((vnode._flags & 4 /* Component */) !== 0) {
                vnode.cref.dispose();
            }
            else if (vnode._children !== null) {
                const children = vnode._children;
                if (typeof children !== "string") {
                    for (let i = 0; i < children.length; i++) {
                        vNodeDispose(children[i]);
                    }
                }
            }
        }
        else {
            vNodeDetach(vnode);
        }
    }
    /**
     * Instantiate a DOM Node or Component from the Virtual DOM Node.
     *
     * This method doesn't set any attributes, or create children, to render internal representation of the virtual node,
     * use `vNodeRender` method.
     */
    function vNodeInstantiate(vnode, owner) {
        const flags = vnode._flags;
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if (vnode.ref !== null) {
                throw new Error("Failed to create VNode: VNode already has a reference to the DOM node.");
            }
        }
        if ((flags & 1 /* Text */) !== 0) {
            vnode.ref = document.createTextNode(vnode._props);
        }
        else if ((flags & 2 /* Element */) !== 0) {
            if ((flags & 131072 /* VModel */) === 0) {
                if ((flags & 32768 /* Svg */) === 0) {
                    vnode.ref = document.createElement(vnode._tag);
                }
                else {
                    vnode.ref = document.createElementNS(SvgNamespace, vnode._tag);
                }
            }
            else {
                vnode.ref = vnode._tag.createElement();
            }
        }
        else if ((flags & 256 /* KeepAlive */) === 0) {
            const c = vnode._tag.createComponent(owner, vnode._props);
            vnode.ref = c.element;
            vnode.cref = c;
        }
    }
    /**
     * Render internal representation of the Virtual DOM Node.
     */
    function vNodeRender(vnode, owner) {
        let i;
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if (vnode.ref === null) {
                throw new Error("Failed to render VNode: VNode should be created before render.");
            }
            if ((vnode._flags & 8 /* Root */) !== 0 && owner === undefined) {
                throw new Error("Failed to render VNode: VNode component root should have an owner.");
            }
            if ((vnode._debugProperties.flags & 1 /* Rendered */) !== 0) {
                throw new Error("Failed to render VNode: VNode cannot be rendered twice.");
            }
            if ((vnode._debugProperties.flags & 2 /* Mounted */) !== 0) {
                throw new Error("Failed to render VNode: VNode cannot be rendered after mount.");
            }
            vnode._debugProperties.flags |= 2 /* Mounted */;
        }
        let il;
        let key;
        let keys;
        const flags = vnode._flags;
        let ref;
        if ((flags & (2 /* Element */ | 8 /* Root */)) !== 0) {
            ref = vnode.ref;
            if ((flags & 262144 /* VModelUpdateHandler */) === 0) {
                const props = vnode._props;
                if (props !== null) {
                    keys = Object.keys(props);
                    for (i = 0, il = keys.length; i < il; i++) {
                        key = keys[i];
                        ref[key] = props[key];
                    }
                }
                if (vnode._attrs !== null) {
                    keys = Object.keys(vnode._attrs);
                    for (i = 0, il = keys.length; i < il; i++) {
                        key = keys[i];
                        setAttr(ref, key, vnode._attrs[key]);
                    }
                }
                if (vnode._style !== null) {
                    if ((flags & 32768 /* Svg */) === 0) {
                        ref.style.cssText = vnode._style;
                    }
                    else {
                        ref.setAttribute("style", vnode._style);
                    }
                }
                if (vnode._className !== null) {
                    if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                        const className = ref.getAttribute("class");
                        if ((flags & 8 /* Root */) !== 0 && className) {
                            printError(`VNode render: Component root node overwrited className property` +
                                ` "${className}" with "${vnode._className}".`);
                        }
                    }
                    if ((flags & 32768 /* Svg */) === 0) {
                        ref.className = vnode._className;
                    }
                    else {
                        ref.setAttribute("class", vnode._className);
                    }
                }
            }
            else {
                if ((flags & 8 /* Root */) === 0) {
                    vnode._tag.update(ref, undefined, vnode._props);
                }
                else {
                    owner.descriptor._tag.update(ref, undefined, vnode._props);
                }
            }
            const children = vnode._children;
            if (children !== null) {
                if ((vnode._flags & 2048 /* UnsafeHTML */) === 0) {
                    if ((vnode._flags & 192 /* InputElement */) === 0) {
                        if (typeof children === "string") {
                            ref.textContent = children;
                        }
                        else {
                            for (i = 0, il = children.length; i < il; i++) {
                                vNodeInsertChild(vnode, children[i], null, owner);
                            }
                        }
                    }
                    else {
                        if ((vnode._flags & 64 /* TextInputElement */) !== 0) {
                            vnode.ref.value = vnode._children;
                        }
                        else {
                            vnode.ref.checked = vnode._children;
                        }
                    }
                }
                else {
                    ref.innerHTML = children;
                }
            }
        }
        else if ((flags & 4 /* Component */) !== 0) {
            ref = vnode.ref;
            if (vnode._className !== null) {
                if ((flags & 32768 /* Svg */) === 0) {
                    ref.className = vnode._className;
                }
                else {
                    ref.setAttribute("class", vnode._className);
                }
            }
            updateComponent(vnode.cref);
        }
    }
    function vNodeInsertChild(parent, node, nextRef, owner) {
        const container = parent.ref;
        if (node.ref === null) {
            vNodeInstantiate(node, owner);
            container.insertBefore(node.ref, nextRef);
            vNodeAttached(node);
            vNodeRender(node, owner);
        }
        else {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((node._flags & 256 /* KeepAlive */) === 0) {
                    throw new Error("Failed to replace node: VNode instance already has been used to create DOM node.");
                }
            }
            container.insertBefore(node.ref, nextRef);
            vNodeAttach(node);
            updateComponent(node.cref, (node._flags & 512 /* BindOnce */) === 0 ? node._props : null);
        }
    }
    function vNodeReplaceChild(parent, newNode, refNode, owner) {
        const container = parent.ref;
        if (newNode.ref === null) {
            vNodeInstantiate(newNode, owner);
            container.replaceChild(newNode.ref, refNode.ref);
            vNodeAttached(newNode);
            vNodeRender(newNode, owner);
        }
        else {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((newNode._flags & 256 /* KeepAlive */) === 0) {
                    throw new Error("Failed to replace node: VNode instance already has been used to create DOM node.");
                }
            }
            container.replaceChild(newNode.ref, refNode.ref);
            vNodeAttach(newNode);
            updateComponent(newNode.cref, (newNode._flags & 512 /* BindOnce */) === 0 ? newNode._props : null);
        }
        vNodeDispose(refNode);
    }
    function vNodeMoveChild(parent, node, nextRef) {
        parent.ref.insertBefore(node.ref, nextRef);
    }
    function vNodeRemoveChild(parent, node) {
        parent.ref.removeChild(node.ref);
        vNodeDispose(node);
    }
    /**
     * Sync two VNodes
     *
     * When node `a` is synced with node `b`, `a` node should be considered as destroyed, and any access to it after sync
     * is an undefined behavior.
     */
    function syncVNodes(a, b, owner) {
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if ((a._debugProperties.flags & (1 /* Rendered */ | 2 /* Mounted */)) === 0) {
                throw new Error("Failed to sync VNode: VNode should be rendered or mounted before sync.");
            }
            b._debugProperties.flags |= a._debugProperties.flags &
                (1 /* Rendered */ | 2 /* Mounted */ |
                    4 /* Attached */ | 8 /* Detached */);
        }
        const ref = a.ref;
        const flags = a._flags;
        let component;
        let className;
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if (a._flags !== b._flags) {
                throw new Error(`Failed to sync VNode: flags does not match (old: ${a._flags}, new: ${b._flags}).`);
            }
            if (a._tag !== b._tag) {
                throw new Error(`Failed to sync VNode: tags does not match (old: ${a._tag}, new: ${b._tag}).`);
            }
            if (a._key !== b._key) {
                throw new Error(`Failed to sync VNode: keys does not match (old: ${a._key}, new: ${b._key}).`);
            }
            if (b.ref !== null && a.ref !== b.ref) {
                throw new Error("Failed to sync VNode: reusing VNodes isn't allowed unless it has the same ref.");
            }
        }
        b.ref = a.ref;
        if ((flags & 1 /* Text */) !== 0) {
            if (a._props !== b._props) {
                a.ref.nodeValue = b._props;
            }
        }
        else if ((flags & (2 /* Element */ | 8 /* Root */)) !== 0) {
            if ((flags & 262144 /* VModelUpdateHandler */) === 0) {
                if (a._props !== b._props) {
                    syncStaticShapeProps(ref, a._props, b._props);
                }
                if (a._attrs !== b._attrs) {
                    if ((a._flags & 32 /* DynamicShapeAttrs */) === 0) {
                        syncStaticShapeAttrs(ref, a._attrs, b._attrs);
                    }
                    else {
                        syncDynamicShapeAttrs(ref, a._attrs, b._attrs);
                    }
                }
                if (a._style !== b._style) {
                    const style = (b._style === null) ? "" : b._style;
                    if ((flags & 32768 /* Svg */) === 0) {
                        ref.style.cssText = style;
                    }
                    else {
                        ref.setAttribute("style", style);
                    }
                }
                if (a._className !== b._className) {
                    className = (b._className === null) ? "" : b._className;
                    if ((flags & 32768 /* Svg */) === 0) {
                        ref.className = className;
                    }
                    else {
                        ref.setAttribute("class", className);
                    }
                }
            }
            else if (a._props !== b._props) {
                if ((flags & 8 /* Root */) === 0) {
                    a._tag.update(ref, a._props, b._props);
                }
                else {
                    owner.descriptor._tag.update(ref, a._props, b._props);
                }
            }
            if ((a._flags & 192 /* InputElement */) === 0) {
                if (a._children !== b._children) {
                    if ((a._flags & 2048 /* UnsafeHTML */) === 0) {
                        _syncChildren(a, a._children, b._children, owner);
                    }
                    else {
                        ref.innerHTML = b._children;
                    }
                }
            }
            else {
                if ((flags & 64 /* TextInputElement */) !== 0) {
                    if (ref.value !== b._children) {
                        ref.value = b._children;
                    }
                }
                else {
                    if (ref.checked !== b._children) {
                        ref.checked = b._children;
                    }
                }
            }
        }
        else {
            component = b.cref = a.cref;
            if (a._className !== b._className) {
                className = (b._className === null) ? "" : b._className;
                if ((flags & 32768 /* Svg */) === 0) {
                    ref.className = className;
                }
                else {
                    ref.setAttribute("class", className);
                }
            }
            if (((flags & 1024 /* ImmutableProps */) === 0) || a._props !== b._props) {
                updateComponent(component, (flags & 512 /* BindOnce */) === 0 ? b._props : undefined);
            }
        }
    }
    /**
     * Check if two nodes can be synced.
     *
     * Two nodes can be synced when their flags, tags and keys are identical.
     */
    function _canSyncVNodes(a, b) {
        return (a._flags === b._flags &&
            a._tag === b._tag &&
            a._key === b._key);
    }
    /**
     * Sync old children list with the new one.
     */
    function _syncChildren(parent, a, b, owner) {
        let aNode;
        let bNode;
        let i = 0;
        let synced = false;
        if (typeof a === "string") {
            if (b === null) {
                parent.ref.removeChild(parent.ref.firstChild);
            }
            else if (typeof b === "string") {
                const c = parent.ref.firstChild;
                if (c) {
                    c.nodeValue = b;
                }
                else {
                    parent.ref.textContent = b;
                }
            }
            else {
                parent.ref.removeChild(parent.ref.firstChild);
                while (i < b.length) {
                    vNodeInsertChild(parent, b[i++], null, owner);
                }
            }
        }
        else if (typeof b === "string") {
            if (a !== null) {
                parent.ref.textContent = "";
                while (i < a.length) {
                    vNodeDispose(a[i++]);
                }
            }
            parent.ref.textContent = b;
        }
        else {
            if (a !== null && a.length !== 0) {
                if (b === null || b.length === 0) {
                    // b is empty, remove all children from a.
                    parent.ref.textContent = "";
                    while (i < a.length) {
                        vNodeDispose(a[i++]);
                    }
                }
                else {
                    if (a.length === 1 && b.length === 1) {
                        // Fast path when a and b have only one child.
                        aNode = a[0];
                        bNode = b[0];
                        if (_canSyncVNodes(aNode, bNode)) {
                            syncVNodes(aNode, bNode, owner);
                        }
                        else {
                            vNodeReplaceChild(parent, bNode, aNode, owner);
                        }
                    }
                    else if (a.length === 1) {
                        // Fast path when a have 1 child.
                        aNode = a[0];
                        if ((parent._flags & 16 /* TrackByKeyChildren */) === 0) {
                            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                                if ((parent._debugProperties.flags & 32 /* DisabledChildrenShapeError */) === 0) {
                                    printError("VNode sync children: children shape is changing, you should enable tracking by key with " +
                                        "VNode method trackByKeyChildren(children).\n" +
                                        "If you certain that children shape changes won't cause any problems with losing " +
                                        "state, you can remove this error message with VNode method disableChildrenShapeError().");
                                }
                            }
                            while (i < b.length) {
                                bNode = b[i++];
                                if (_canSyncVNodes(aNode, bNode)) {
                                    syncVNodes(aNode, bNode, owner);
                                    synced = true;
                                    break;
                                }
                                vNodeInsertChild(parent, bNode, aNode.ref, owner);
                            }
                        }
                        else {
                            while (i < b.length) {
                                bNode = b[i++];
                                if (aNode._key === bNode._key) {
                                    if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                                        if (!_canSyncVNodes(aNode, bNode)) {
                                            throw new Error("VNode sync children failed: cannot sync two different children with the" +
                                                " same key.");
                                        }
                                    }
                                    syncVNodes(aNode, bNode, owner);
                                    synced = true;
                                    break;
                                }
                                vNodeInsertChild(parent, bNode, aNode.ref, owner);
                            }
                        }
                        if (synced) {
                            while (i < b.length) {
                                vNodeInsertChild(parent, b[i++], null, owner);
                            }
                        }
                        else {
                            vNodeRemoveChild(parent, aNode);
                        }
                    }
                    else if (b.length === 1) {
                        // Fast path when b have 1 child.
                        bNode = b[0];
                        if ((parent._flags & 16 /* TrackByKeyChildren */) === 0) {
                            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                                if ((parent._debugProperties.flags & 32 /* DisabledChildrenShapeError */) === 0) {
                                    printError("VNode sync children: children shape is changing, you should enable tracking by key with " +
                                        "VNode method trackByKeyChildren(children).\n" +
                                        "If you certain that children shape changes won't cause any problems with losing " +
                                        "state, you can remove this error message with VNode method disableChildrenShapeError().");
                                }
                            }
                            while (i < a.length) {
                                aNode = a[i++];
                                if (_canSyncVNodes(aNode, bNode)) {
                                    syncVNodes(aNode, bNode, owner);
                                    synced = true;
                                    break;
                                }
                                vNodeRemoveChild(parent, aNode);
                            }
                        }
                        else {
                            while (i < a.length) {
                                aNode = a[i++];
                                if (aNode._key === bNode._key) {
                                    if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                                        if (!_canSyncVNodes(aNode, bNode)) {
                                            throw new Error("VNode sync children failed: cannot sync two different children with the" +
                                                " same key.");
                                        }
                                    }
                                    syncVNodes(aNode, bNode, owner);
                                    synced = true;
                                    break;
                                }
                                vNodeRemoveChild(parent, aNode);
                            }
                        }
                        if (synced) {
                            while (i < a.length) {
                                vNodeRemoveChild(parent, a[i++]);
                            }
                        }
                        else {
                            vNodeInsertChild(parent, bNode, null, owner);
                        }
                    }
                    else {
                        // a and b have more than 1 child.
                        if ((parent._flags & 16 /* TrackByKeyChildren */) === 0) {
                            _syncChildrenNaive(parent, a, b, owner);
                        }
                        else {
                            _syncChildrenTrackByKeys(parent, a, b, owner);
                        }
                    }
                }
            }
            else if (b !== null && b.length > 0) {
                // a is empty, insert all children from b.
                for (i = 0; i < b.length; i++) {
                    vNodeInsertChild(parent, b[i], null, owner);
                }
            }
        }
    }
    /**
     * Sync children naive way.
     *
     * Any heuristics that is used in this algorithm is an undefined behaviour, and external dependencies should not rely on
     * any knowledge about this algorithm, because it can be changed in any time.
     *
     * This naive algorithm is quite simple:
     *
     *  A: -> [a a c d e g g] <-
     *  B: -> [a a f d c g] <-
     *
     * It starts by iterating over old children list `A` and new children list `B` from both ends.
     *
     *  A: -> [a b c d e g g] <-
     *  B: -> [a b f d c g] <-
     *
     * When it find nodes that have the same key, tag and flags, it will sync them. Node "a" and "b" on the right side, and
     * node "g" on the right side will be synced.
     *
     *  A: -> [c d e g]
     *  B: -> [f d c]
     *
     * Then it start iterating over old and new children lists from the left side and check if nodes can be synced. Nodes
     * "c" and "f" can't be synced, remove node "c" and insert new node "f".
     *
     *  A: -> [d e g]
     *  B: -> [d c]
     *
     * Node "d" is synced.
     *
     *  A: -> [e g]
     *  B: -> [c]
     *
     * Node "e" removed, node "c" inserted.
     *
     *  A: -> [g]
     *  B:    []
     *
     * Length of the old list is larger than length of the new list, remove remaining nodes from the old list.
     *
     */
    function _syncChildrenNaive(parent, a, b, owner) {
        let aStart = 0;
        let bStart = 0;
        let aEnd = a.length - 1;
        let bEnd = b.length - 1;
        let aNode;
        let bNode;
        let nextPos;
        let next;
        // Sync similar nodes at the beginning.
        while (aStart <= aEnd && bStart <= bEnd) {
            aNode = a[aStart];
            bNode = b[bStart];
            if (!_canSyncVNodes(aNode, bNode)) {
                break;
            }
            aStart++;
            bStart++;
            syncVNodes(aNode, bNode, owner);
        }
        // Sync similar nodes at the end.
        while (aStart <= aEnd && bStart <= bEnd) {
            aNode = a[aEnd];
            bNode = b[bEnd];
            if (!_canSyncVNodes(aNode, bNode)) {
                break;
            }
            aEnd--;
            bEnd--;
            syncVNodes(aNode, bNode, owner);
        }
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if ((aStart <= aEnd || bStart <= bEnd) &&
                ((parent._debugProperties.flags & 32 /* DisabledChildrenShapeError */) === 0)) {
                printError("VNode sync children: children shape is changing, you should enable tracking by key with " +
                    "VNode method trackByKeyChildren(children).\n" +
                    "If you certain that children shape changes won't cause any problems with losing " +
                    "state, you can remove this error message with VNode method disableChildrenShapeError().");
            }
        }
        // Iterate over the remaining nodes and if they have the same type, then sync, otherwise just
        // remove the old node and insert the new one.
        while (aStart <= aEnd && bStart <= bEnd) {
            aNode = a[aStart++];
            bNode = b[bStart++];
            if (_canSyncVNodes(aNode, bNode)) {
                syncVNodes(aNode, bNode, owner);
            }
            else {
                vNodeReplaceChild(parent, bNode, aNode, owner);
            }
        }
        if (aStart <= aEnd) {
            // All nodes from a are synced, remove the rest.
            do {
                vNodeRemoveChild(parent, a[aStart++]);
            } while (aStart <= aEnd);
        }
        else if (bStart <= bEnd) {
            // All nodes from b are synced, insert the rest.
            nextPos = bEnd + 1;
            next = nextPos < b.length ? b[nextPos].ref : null;
            do {
                vNodeInsertChild(parent, b[bStart++], next, owner);
            } while (bStart <= bEnd);
        }
    }
    /**
     * Sync children with track by keys algorithm.
     *
     * This algorithm finds a minimum number of DOM operations. It works in several steps:
     *
     * 1. Find common suffix and prefix, and perform simple moves on the edges.
     *
     * This optimization technique is searching for nodes with identical keys by simultaneously iterating over nodes in the
     * old children list `A` and new children list `B` from both sides:
     *
     *  A: -> [a b c d e f g] <-
     *  B: -> [a b f d c g] <-
     *
     * Here we can skip nodes "a" and "b" at the begininng, and node "g" at the end.
     *
     *  A: -> [c d e f] <-
     *  B: -> [f d c] <-
     *
     * At this position it will try to look at the opposite edge, and if there is a node with the same key at the opposite
     * edge, it will perform simple move operation. Node "c" is moved to the right edge, and node "f" is moved to the left
     * edge.
     *
     *  A: -> [d e] <-
     *  B: -> [d] <-
     *
     * Now it will try again to find common prefix and suffix, node "d" is the same, so we can skip it.
     *
     *  A: [e]
     *  B: []
     *
     * Here it will check if the size of one of the list is equal to zero, and if length of the old children list is zero,
     * it will insert all remaining nodes from the new list, or if length of the new children list is zero, it will remove
     * all remaining nodes from the old list.
     *
     * This simple optimization technique will cover most of the real world use cases, even reversing the children list,
     * except for sorting.
     *
     * When algorithm couldn't find a solution with this simple optimization technique, it will go to the next step of the
     * algorithm. For example:
     *
     *  A: -> [a b c d e f g] <-
     *  B: -> [a c b h f e g] <-
     *
     * Nodes "a" and "g" at the edges are the same, skipping them.
     *
     *  A: -> [b c d e f] <-
     *  B: -> [c b h f e] <-
     *
     * Here we are stuck, so we need to switch to the next step.
     *
     * 2. Look for removed and inserted nodes, and simultaneously check if one of the nodes is moved.
     *
     * First we create an array `P` with the length of the new children list and assign to each position value `-1`, it has
     * a meaning of a new node that should be inserted. Later we will assign node positions in the old children list to this
     * array.
     *
     *  A: [b c d e f]
     *  B: [c b h f e]
     *  P: [. . . . .] // . == -1
     *
     * Then we need to build an index `I` that maps keys with node positions of the remaining nodes from the new children
     * list.
     *
     *  A: [b c d e f]
     *  B: [c b h f e]
     *  P: [. . . . .] // . == -1
     *  I: {
     *    c: 0,
     *    b: 1,
     *    h: 2,
     *    f: 3,
     *    e: 4,
     *  }
     *  last = 0
     *
     * With this index, we start to iterate over the remaining nodes from the old children list and check if we can find a
     * node with the same key in the index. If we can't find any node, it means that it should be removed, otherwise we
     * assign position of the node in the old children list to the positions array.
     *
     *  A: [b c d e f]
     *      ^
     *  B: [c b h f e]
     *  P: [. 0 . . .] // . == -1
     *  I: {
     *    c: 0,
     *    b: 1, <-
     *    h: 2,
     *    f: 3,
     *    e: 4,
     *  }
     *  last = 1
     *
     * When we assigning positions to the positions array, we also keep a position of the last seen node in the new children
     * list, if the last seen position is larger than current position of the node at the new list, then we are switching
     * `moved` flag to `true`.
     *
     *  A: [b c d e f]
     *        ^
     *  B: [c b h f e]
     *  P: [1 0 . . .] // . == -1
     *  I: {
     *    c: 0, <-
     *    b: 1,
     *    h: 2,
     *    f: 3,
     *    e: 4,
     *  }
     *  last = 1 // last > 0; moved = true
     *
     * The last position `1` is larger than current position of the node at the new list `0`, switching `moved` flag to
     * `true`.
     *
     *  A: [b c d e f]
     *          ^
     *  B: [c b h f e]
     *  P: [1 0 . . .] // . == -1
     *  I: {
     *    c: 0,
     *    b: 1,
     *    h: 2,
     *    f: 3,
     *    e: 4,
     *  }
     *  moved = true
     *
     * Node with key "d" doesn't exist in the index, removing node.
     *
     *  A: [b c d e f]
     *            ^
     *  B: [c b h f e]
     *  P: [1 0 . . 3] // . == -1
     *  I: {
     *    c: 0,
     *    b: 1,
     *    h: 2,
     *    f: 3,
     *    e: 4, <-
     *  }
     *  moved = true
     *
     * Assign position for `e`.
     *
     *  A: [b c d e f]
     *              ^
     *  B: [c b h f e]
     *  P: [1 0 . 4 3] // . == -1
     *  I: {
     *    c: 0,
     *    b: 1,
     *    h: 2,
     *    f: 3, <-
     *    e: 4,
     *  }
     *  moved = true
     *
     * Assign position for 'f'.
     *
     * At this point we are checking if `moved` flag is on, or if the length of the old children list minus the number of
     * removed nodes isn't equal to the length of the new children list. If any of this conditions is true, then we are
     * going to the next step.
     *
     * 3. Find minimum number of moves if `moved` flag is on, or insert new nodes if the length is changed.
     *
     * When `moved` flag is on, we need to find the
     * [longest increasing subsequence](http://en.wikipedia.org/wiki/Longest_increasing_subsequence) in the positions array,
     * and move all nodes that doesn't belong to this subsequence.
     *
     *  A: [b c d e f]
     *  B: [c b h f e]
     *  P: [1 0 . 4 3] // . == -1
     *  LIS:     [1 4]
     *  moved = true
     *
     * Now we just need to simultaneously iterate over the new children list and LIS from the end and check if the current
     * position is equal to a value from LIS.
     *
     *  A: [b c d e f]
     *  B: [c b h f e]
     *              ^  // new_pos == 4
     *  P: [1 0 . 4 3] // . == -1
     *  LIS:     [1 4]
     *              ^  // new_pos == 4
     *  moved = true
     *
     * Node "e" stays at the same place.
     *
     *  A: [b c d e f]
     *  B: [c b h f e]
     *            ^    // new_pos == 3
     *  P: [1 0 . 4 3] // . == -1
     *  LIS:     [1 4]
     *            ^    // new_pos != 1
     *  moved = true
     *
     * Node "f" is moved, move it before the next node "e".
     *
     *  A: [b c d e f]
     *  B: [c b h f e]
     *          ^      // new_pos == 2
     *  P: [1 0 . 4 3] // . == -1
     *          ^      // old_pos == -1
     *  LIS:     [1 4]
     *            ^
     *  moved = true
     *
     * Node "h" has a `-1` value in the positions array, insert new node "h".
     *
     *  A: [b c d e f]
     *  B: [c b h f e]
     *        ^        // new_pos == 1
     *  P: [1 0 . 4 3] // . == -1
     *  LIS:     [1 4]
     *            ^    // new_pos == 1
     *  moved = true
     *
     * Node "b" stays at the same place.
     *
     *  A: [b c d e f]
     *  B: [c b h f e]
     *      ^          // new_pos == 0
     *  P: [1 0 . 4 3] // . == -1
     *  LIS:     [1 4]
     *          ^      // new_pos != undefined
     *  moved = true
     *
     * Node "c" is moved, move it before the next node "b".
     *
     * When moved flag is off, we don't need to find LIS, and we just iterate over the new children list and check its
     * current position in the positions array, if it is `-1`, then we insert new node.
     *
     * That is how children reconciliation algorithm is working in one of the fastest virtual dom libraries :)
     */
    function _syncChildrenTrackByKeys(parent, a, b, owner) {
        let aStart = 0;
        let bStart = 0;
        let aEnd = a.length - 1;
        let bEnd = b.length - 1;
        let aStartNode = a[aStart];
        let bStartNode = b[bStart];
        let aEndNode = a[aEnd];
        let bEndNode = b[bEnd];
        let i;
        let j;
        let nextPos;
        let next;
        let aNode;
        let bNode;
        let node;
        // Step 1
        outer: do {
            // Sync nodes with the same key at the beginning.
            while (aStartNode._key === bStartNode._key) {
                if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                    if (!_canSyncVNodes(aStartNode, bStartNode)) {
                        throw new Error("VNode sync children failed: cannot sync two different children with the same key.");
                    }
                }
                syncVNodes(aStartNode, bStartNode, owner);
                aStart++;
                bStart++;
                if (aStart > aEnd || bStart > bEnd) {
                    break outer;
                }
                aStartNode = a[aStart];
                bStartNode = b[bStart];
            }
            // Sync nodes with the same key at the end.
            while (aEndNode._key === bEndNode._key) {
                if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                    if (!_canSyncVNodes(aEndNode, bEndNode)) {
                        throw new Error("VNode sync children failed: cannot sync two different children with the same key.");
                    }
                }
                syncVNodes(aEndNode, bEndNode, owner);
                aEnd--;
                bEnd--;
                if (aStart > aEnd || bStart > bEnd) {
                    break outer;
                }
                aEndNode = a[aEnd];
                bEndNode = b[bEnd];
            }
            // Move and sync nodes from left to right.
            if (aStartNode._key === bEndNode._key) {
                if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                    if (!_canSyncVNodes(aStartNode, bEndNode)) {
                        throw new Error("VNode sync children failed: cannot sync two different children with the same key.");
                    }
                }
                syncVNodes(aStartNode, bEndNode, owner);
                nextPos = bEnd + 1;
                next = nextPos < b.length ? b[nextPos].ref : null;
                vNodeMoveChild(parent, bEndNode, next);
                aStart++;
                bEnd--;
                if (aStart > aEnd || bStart > bEnd) {
                    break outer;
                }
                aStartNode = a[aStart];
                bEndNode = b[bEnd];
                // In a real-world scenarios there is a higher chance that next node after the move will be the same, so we
                // immediately jump to the start of this prefix/suffix algo.
                continue outer;
            }
            // Move and sync nodes from right to left.
            if (aEndNode._key === bStartNode._key) {
                if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                    if (!_canSyncVNodes(aEndNode, bStartNode)) {
                        throw new Error("VNode sync children failed: cannot sync two different children with the same key.");
                    }
                }
                syncVNodes(aEndNode, bStartNode, owner);
                vNodeMoveChild(parent, bStartNode, aStartNode.ref);
                aEnd--;
                bStart++;
                if (aStart > aEnd || bStart > bEnd) {
                    break outer;
                }
                aEndNode = a[aEnd];
                bStartNode = b[bStart];
                continue outer;
            }
        } while (false);
        if (aStart > aEnd) {
            // All nodes from a are synced, insert the rest from b.
            nextPos = bEnd + 1;
            next = nextPos < b.length ? b[nextPos].ref : null;
            while (bStart <= bEnd) {
                vNodeInsertChild(parent, b[bStart++], next, owner);
            }
        }
        else if (bStart > bEnd) {
            // All nodes from b are synced, remove the rest from a.
            while (aStart <= aEnd) {
                vNodeRemoveChild(parent, a[aStart++]);
            }
        }
        else {
            let aLength = aEnd - aStart + 1;
            let bLength = bEnd - bStart + 1;
            const sources = new Array(bLength);
            // Mark all nodes as inserted.
            for (i = 0; i < bLength; i++) {
                sources[i] = -1;
            }
            let moved = false;
            let removed;
            let pos = 0;
            let removeOffset = 0;
            let synced = 0;
            // When children lists are small, we are using naive O(N) algorithm to find if child is removed.
            if ((bLength <= 4) || ((aLength * bLength) <= 16)) {
                for (i = aStart; i <= aEnd; i++) {
                    removed = true;
                    aNode = a[i];
                    if (synced < bLength) {
                        for (j = bStart; j <= bEnd; j++) {
                            bNode = b[j];
                            if (aNode._key === bNode._key) {
                                sources[j - bStart] = i;
                                if (pos > j) {
                                    moved = true;
                                }
                                else {
                                    pos = j;
                                }
                                if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                                    if (!_canSyncVNodes(aNode, bNode)) {
                                        throw new Error("VNode sync children failed: cannot sync two different children with the same key.");
                                    }
                                }
                                syncVNodes(aNode, bNode, owner);
                                synced++;
                                removed = false;
                                break;
                            }
                        }
                    }
                    if (removed) {
                        vNodeRemoveChild(parent, aNode);
                        removeOffset++;
                    }
                }
            }
            else {
                const keyIndex = new Map();
                for (i = bStart; i <= bEnd; i++) {
                    node = b[i];
                    keyIndex.set(node._key, i);
                }
                for (i = aStart; i <= aEnd; i++) {
                    removed = true;
                    aNode = a[i];
                    if (synced < bLength) {
                        j = keyIndex.get(aNode._key);
                        if (j !== undefined) {
                            bNode = b[j];
                            sources[j - bStart] = i;
                            if (pos > j) {
                                moved = true;
                            }
                            else {
                                pos = j;
                            }
                            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                                if (!_canSyncVNodes(aNode, bNode)) {
                                    throw new Error("VNode sync children failed: cannot sync two different children with the same key.");
                                }
                            }
                            syncVNodes(aNode, bNode, owner);
                            synced++;
                            removed = false;
                        }
                    }
                    if (removed) {
                        vNodeRemoveChild(parent, aNode);
                        removeOffset++;
                    }
                }
            }
            // Step 3
            if (moved) {
                const seq = _lis(sources);
                j = seq.length - 1;
                for (i = bLength - 1; i >= 0; i--) {
                    if (sources[i] === -1) {
                        pos = i + bStart;
                        node = b[pos];
                        nextPos = pos + 1;
                        next = nextPos < b.length ? b[nextPos].ref : null;
                        vNodeInsertChild(parent, node, next, owner);
                    }
                    else {
                        if (j < 0 || i !== seq[j]) {
                            pos = i + bStart;
                            node = b[pos];
                            nextPos = pos + 1;
                            next = nextPos < b.length ? b[nextPos].ref : null;
                            vNodeMoveChild(parent, node, next);
                        }
                        else {
                            j--;
                        }
                    }
                }
            }
            else if (aLength - removeOffset !== bLength) {
                for (i = bLength - 1; i >= 0; i--) {
                    if (sources[i] === -1) {
                        pos = i + bStart;
                        node = b[pos];
                        nextPos = pos + 1;
                        next = nextPos < b.length ? b[nextPos].ref : null;
                        vNodeInsertChild(parent, node, next, owner);
                    }
                }
            }
        }
    }
    /**
     * Slightly modified Longest Increased Subsequence algorithm, it ignores items that have -1 value, they're representing
     * new items.
     *
     * http://en.wikipedia.org/wiki/Longest_increasing_subsequence
     */
    function _lis(a) {
        const p = a.slice(0);
        const result = [];
        result.push(0);
        let u;
        let v;
        for (let i = 0, il = a.length; i < il; i++) {
            if (a[i] === -1) {
                continue;
            }
            let j = result[result.length - 1];
            if (a[j] < a[i]) {
                p[i] = j;
                result.push(i);
                continue;
            }
            u = 0;
            v = result.length - 1;
            while (u < v) {
                let c = ((u + v) / 2) | 0;
                if (a[result[c]] < a[i]) {
                    u = c + 1;
                }
                else {
                    v = c;
                }
            }
            if (a[i] < a[result[u]]) {
                if (u > 0) {
                    p[i] = result[u - 1];
                }
                result[u] = i;
            }
        }
        u = result.length;
        v = result[u - 1];
        while (u-- > 0) {
            result[u] = v;
            v = p[v];
        }
        return result;
    }
    /**
     * Sync HTML attributes with static shape.
     *
     * Attributes with static shape should have the same keys.
     *
     * Valid:
     *
     *   a: { title: "Google", href: "https://www.google.com" }
     *   b: { title: "Facebook", href: "https://www.facebook.com" }
     *
     * Invalid:
     *
     *  a: { title: "Google", href: "https://www.google.com" }
     *  b: { title: "Facebook" }
     */
    function syncStaticShapeAttrs(node, a, b) {
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if (a === null || b === null) {
                throw new Error("Failed to update attrs with static shape: attrs object have dynamic shape.");
            }
        }
        let keys = Object.keys(a);
        let key;
        let i;
        for (i = 0; i < keys.length; i++) {
            key = keys[i];
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if (!b.hasOwnProperty(key)) {
                    throw new Error("Failed to update attrs with static shape: attrs object have dynamic shape.");
                }
            }
            const bValue = b[key];
            if (a[key] !== bValue) {
                setAttr(node, key, bValue);
            }
        }
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            keys = Object.keys(b);
            for (i = 0; i < keys.length; i++) {
                key = keys[i];
                if (!a.hasOwnProperty(key)) {
                    throw new Error("Failed to update attrs with static shape: attrs object have dynamic shape.");
                }
            }
        }
    }
    /**
     * Sync HTML attributes with dynamic shape.
     *
     * Attributes with dynamic shape can have any keys, missing keys will be removed with `node.removeAttribute` method.
     *
     *   a: { title: "Google", href: "https://www.google.com" }
     *   b: { title: "Google" }
     *
     * In this example `href` attribute will be removed.
     */
    function syncDynamicShapeAttrs(node, a, b) {
        let i;
        let keys;
        let key;
        if (a !== null) {
            if (b === null) {
                // b is empty, remove all attributes from a.
                keys = Object.keys(a);
                for (i = 0; i < keys.length; i++) {
                    node.removeAttribute(keys[i]);
                }
            }
            else {
                // Remove and update attributes.
                keys = Object.keys(a);
                for (i = 0; i < keys.length; i++) {
                    key = keys[i];
                    if (b.hasOwnProperty(key)) {
                        const bValue = b[key];
                        if (a[key] !== bValue) {
                            setAttr(node, key, bValue);
                        }
                    }
                    else {
                        node.removeAttribute(key);
                    }
                }
                // Insert new attributes.
                keys = Object.keys(b);
                for (i = 0; i < keys.length; i++) {
                    key = keys[i];
                    if (!a.hasOwnProperty(key)) {
                        setAttr(node, key, b[key]);
                    }
                }
            }
        }
        else if (b !== null) {
            // a is empty, insert all attributes from b.
            keys = Object.keys(b);
            for (i = 0; i < keys.length; i++) {
                key = keys[i];
                setAttr(node, key, b[key]);
            }
        }
    }
    /**
     * Sync HTML properties with static shape.
     *
     * Properties with static shape should have the same keys.
     *
     * Valid:
     *
     *   a: { title: "Google", href: "https://www.google.com" }
     *   b: { title: "Facebook", href: "https://www.facebook.com" }
     *
     * Invalid:
     *
     *  a: { title: "Google", href: "https://www.google.com" }
     *  b: { title: "Facebook" }
     */
    function syncStaticShapeProps(node, a, b) {
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if (a === null || b === null) {
                throw new Error("Failed to update props with static shape: props object have dynamic shape.");
            }
        }
        let keys = Object.keys(a);
        let key;
        let i;
        for (i = 0; i < keys.length; i++) {
            key = keys[i];
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if (!b.hasOwnProperty(key)) {
                    throw new Error("Failed to update props with static shape: props object have dynamic shape.");
                }
            }
            const bValue = b[key];
            if (a[key] !== bValue) {
                node[key] = bValue;
            }
        }
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            keys = Object.keys(b);
            for (i = 0; i < keys.length; i++) {
                key = keys[i];
                if (!a.hasOwnProperty(key)) {
                    throw new Error("Failed to update attrs with static shape: attrs object have dynamic shape.");
                }
            }
        }
    }
    /**
     * Create a VNode representing a [Text] node.
     */
    function createVText(content) {
        return new VNode(1 /* Text */, null, content);
    }
    /**
     * Create a VNode representing an [Element] node.
     */
    function createVElement(tagName) {
        return new VNode(2 /* Element */, tagName, null);
    }
    /**
     * Create a VNode representing a [SVGElement] node.
     */
    function createVSvgElement(tagName) {
        return new VNode(2 /* Element */ | 32768 /* Svg */, tagName, null);
    }
    /**
     * Create a VNode representing a Component root node.
     */
    function createVRoot() {
        return new VNode(8 /* Root */, null, null);
    }

    /**
     * Component Descriptor registry used in DEBUG mode.
     */
    const ComponentDescriptorRegistry = new Map();
    /**
     * Component Descriptor.
     *
     * Each component should declare its properties and behavior in `ComponentDescriptor` object.
     *
     * Component descriptor has two parametric types: first parametric type `P` is a props type and second type `S` is a
     * state type.
     *
     * Component descriptor provides a `createComponent` and `createRootComponent` methods to create component instances.
     *
     * Example:
     *
     *     class State {
     *       xy: number;
     *
     *       constructor(props: number) {
     *         this.xy = props * props;
     *       }
     *     }
     *
     *     const MyComponent = new ComponentDescriptor<number, State>()
     *       .canvas()
     *       .createState((c, props) => new State(props))
     *       .update((c, props, state) => {
     *         const ctx = c.get2DContext();
     *         ctx.fillStyle = 'rgba(0, 0, 0, 1)';
     *         ctx.fillRect(props, props, state.xy, state.xy);
     *       });
     *
     *       const componentInstance = MyComponent.createRootComponent(10);
     *
     * @final
     */
    class ComponentDescriptor {
        constructor(name) {
            this._markFlags = 4 /* Dirty */;
            this._flags = 0;
            this._tag = "div";
            this._newPropsReceived = null;
            this._update = null;
            this._createState = null;
            this._init = null;
            this._attached = null;
            this._detached = null;
            this._disposed = null;
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if (name === undefined) {
                    this.name = "unnamed";
                }
                else {
                    this.name = name;
                    if (ComponentDescriptorRegistry.has(name)) {
                        printError(`Component with name ${name} is already registered in ComponentDescriptorRegistry.`);
                    }
                    else {
                        ComponentDescriptorRegistry.set(name, this);
                    }
                }
            }
            if ("<@KIVI_COMPONENT_RECYCLING@>" === "COMPONENT_RECYCLING_ENABLED") {}
        }
        /**
         * Set tag name for the root element.
         *
         *     const MyComponent = new ComponentDescriptor()
         *       .tagName("table");
         */
        tagName(tagName) {
            this._tag = tagName;
            return this;
        }
        /**
         * Use SVG Namespace to create a root element.
         *
         *     const MySvgComponent = new ComponentDescriptor()
         *       .svg()
         *       .tagName("circle");
         */
        svg() {
            this._markFlags |= 32768 /* Svg */;
            this._flags |= 32768 /* Svg */;
            return this;
        }
        /**
         * Set VModel for the root element.
         *
         *     const MyComponent = new ComponentDescriptor()
         *       .vModel(new VModel("div").attrs({"id": "model"}))
         *       .update((c) => { c.vSync(c.createVRoot().children("content")); });
         */
        vModel(model) {
            this._markFlags |= model._markFlags;
            this._flags |= model._markFlags;
            this._tag = model;
            return this;
        }
        /**
         * Turn component into a canvas.
         *
         *     const MyCanvasComponent = new ComponentDescriptor()
         *       .canvas()
         *       .update((c) => {
         *         const ctx = c.get2DContext();
         *         ctx.fillStyle = "red";
         *         ctx.fillRect(0, 0, 10, 10);
         *       });
         */
        canvas() {
            this._markFlags |= 65536 /* Canvas2D */;
            this._flags |= 65536 /* Canvas2D */;
            this._tag = "canvas";
            return this;
        }
        /**
         * Set lifecycle handler createState.
         *
         * Create state handler should return a new state. It is invoked immediately after component instantiation.
         *
         *     const MyComponent = new ComponentDescriptor<number, number>()
         *       .createState((c, props) => props * props)
         *       .update((c, props, state) => {
         *         c.vSync(c.createVRoot().children(state.toString()));
         *       });
         */
        createState(createState) {
            this._createState = createState;
            return this;
        }
        /**
         * Set lifecycle handler init.
         *
         * `element` and `props` properties will be initialized before init handler is invoked.
         *
         *     const MyComponent = new ComponentDescriptor()
         *       .tagName("button")
         *       .init((c) => {
         *         c.element.addEventListener("click", (e) => {
         *           e.preventDefault();
         *           e.stopPropagation();
         *           console.log("clicked");
         *         });
         *       })
         *       .update((c) => {
         *         c.vSync(c.createVRoot().children("click me"));
         *       });
         */
        init(init) {
            this._init = init;
            return this;
        }
        /**
         * Set newPropsReceived handler.
         *
         * New props received handler overrides default props received behavior and it should mark component as dirty if new
         * received props will cause change in component's representation.
         *
         *     const MyComponent = new ComponentDescriptor<{a: number}, void>()
         *       .newPropsReceived((c, oldProps, newProps) => {
         *         if (oldProps.a !== newProps.a) {
         *           c.markDirty();
         *         }
         *       })
         *       .update((c, props) => {
         *         c.vSync(c.createVRoot().children(c.props.toString()));
         *       });
         */
        newPropsReceived(newPropsReceived) {
            this._newPropsReceived = newPropsReceived;
            return this;
        }
        /**
         * Set lifecycle handler update.
         *
         * Update handler overrides default update behavior.
         *
         *     const MyComponent = new ComponentDescriptor<{a: number}, void>()
         *       .update((c) => {
         *         c.vSync(c.createVRoot().children("content"));
         *       });
         */
        update(update) {
            this._update = update;
            return this;
        }
        /**
         * Set lifecycle handler attached.
         *
         * Attached handler will be invoked when component is attached to the document.
         *
         *     const onChange = new Invalidator();
         *
         *     const MyComponent = new ComponentDescriptor()
         *       .attached((c) => {
         *         c.subscribe(onChange);
         *       })
         *       .update((c) => {
         *         c.vSync(c.createVRoot().children("content"));
         *       });
         */
        attached(attached) {
            this._attached = attached;
            return this;
        }
        /**
         * Set lifecycle handler detached.
         *
         * Detached handler will be invoked when component is detached from the document.
         *
         *     const MyComponent = new ComponentDescriptor<any, {onResize: (e: Event) => void}>()
         *       .createState((c) => ({onResize: (e) => { console.log("window resized"); }})
         *       .attached((c, props, state) => {
         *         window.addEventListener("resize", state.onResize);
         *       })
         *       .detached((c, props, state) => {
         *         window.removeEventListener(state.onResize);
         *       })
         *       .update((c) => {
         *         c.vSync(c.createVRoot().children("content"));
         *       });
         */
        detached(detached) {
            this._detached = detached;
            return this;
        }
        /**
         * Set lifecycle handler disposed.
         *
         * Disposed handler will be invoked when component is disposed.
         *
         *     let allocatedComponents = 0;
         *
         *     const MyComponent = new ComponentDescriptor()
         *       .init((c) => {
         *         allocatedComponents++;
         *       })
         *       .disposed((c) => {
         *         allocatedComponents--;
         *       })
         *       .update((c) => {
         *         c.vSync(c.createVRoot().children("content"));
         *       });
         */
        disposed(disposed) {
            this._disposed = disposed;
            return this;
        }
        /**
         * Enable back reference from DOM element to component instance.
         *
         * Back reference will be assigned to `xtag` property.
         *
         *     const MyComponent = new ComponentDescriptor()
         *       .enableBackRef()
         *       .init((c) => {
         *         c.element.addEventListener("click", (e) => {
         *           console.log(getBackRef<Component>(e.target) === c);
         *         });
         *       })
         *       .update((c) => {
         *         c.vSync(c.createVRoot().children("content"));
         *       });
         */
        enableBackRef() {
            this._flags |= 1 /* EnabledBackRef */;
            return this;
        }
        /**
         * Enable component recycling.
         *
         * When component recycling is enabled, components will be instantiated from recycled pool.
         *
         *     const MyComponent = new ComponentDescriptor()
         *       .enableRecycling(100)
         *       .update((c) => {
         *         c.vSync(c.createVRoot().children("content"));
         *       });
         */
        enableRecycling(maxRecycled) {
            if ("<@KIVI_COMPONENT_RECYCLING@>" === "COMPONENT_RECYCLING_ENABLED") {}
            return this;
        }
        /**
         * Component has immutable props.
         */
        immutableProps() {
            this._markFlags |= 1024 /* ImmutableProps */;
            return this;
        }
        /**
         * Create a Virtual DOM node.
         *
         *     const MyComponent = new ComponentDescriptor<number, void>()
         *       .update((c, props) => {
         *         c.vSync(c.createVRoot().children(props.toString()));
         *       });
         *
         *     const vnode = MyComponent.createVNode(10);
         */
        createVNode(props) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._markFlags & 1024 /* ImmutableProps */) !== 0) {
                    throw new Error("Failed to create VNode: VNodes for components with immutable props should be created with " +
                        "createImmutableVNode method.");
                }
            }
            return new VNode(4 /* Component */, this, props === undefined ? null : props);
        }
        /**
         * Create a Virtual DOM node with immutable props.
         *
         * Immutable props will be checked for identity before triggering an update, if props identity is the same, then
         * component won't be updated.
         *
         *     const MyComponent = new ComponentDescriptor<number, void>()
         *       .update((c, props) => {
         *         c.vSync(c.createVRoot().children(props.toString()));
         *       });
         *
         *     const vnode = MyComponent.createImmutableVNode(10);
         */
        createImmutableVNode(props) {
            return new VNode(4 /* Component */ | 1024 /* ImmutableProps */, this, props === undefined ? null : props);
        }
        /**
         * Creates a component instance without parent.
         *
         *     const MyComponent = new ComponentDescriptor<number, void>()
         *       .update((c, props) => {
         *         c.vSync(c.createVRoot().children(props.toString()));
         *       });
         *
         *     const component = MyComponent.createRootComponent(10);
         */
        createRootComponent(props) {
            return this.createComponent(undefined, props);
        }
        /**
         * Create a Component.
         *
         *     const ChildComponent = new ComponentDescriptor()
         *       .update((c) => {
         *         c.vSync(c.createVRoot().children("child"));
         *       });
         *
         *     const ParentComponent = new ComponentDescriptor()
         *       .init((c) => {
         *         const child = ChildComponent.createComponent(c);
         *         c.element.appendChild(child.element);
         *         child.attached();
         *         child.update();
         *       });
         *
         *     const rootComponent = MyComponent.createRootComponent();
         */
        createComponent(parent, props) {
            let element;
            let component;
            if ("<@KIVI_COMPONENT_RECYCLING@>" !== "COMPONENT_RECYCLING_ENABLED" ||
                ((this._flags & 524288 /* EnabledRecycling */) === 0) ||
                (this._recycledPool.length === 0)) {
                if ((this._flags & 131072 /* VModel */) === 0) {
                    element = ((this._flags & 32768 /* Svg */) === 0) ?
                        document.createElement(this._tag) :
                        document.createElementNS(SvgNamespace, this._tag);
                }
                else {
                    element = this._tag.createElement();
                }
                component = new Component(this._markFlags, this, element, parent, props);
                if ((this._flags & 1 /* EnabledBackRef */) !== 0) {
                    element.xtag = component;
                }
                if (this._createState !== null) {
                    component.state = this._createState(component, component.props);
                }
                if (this._init !== null) {
                    this._init(component, component.props, component.state);
                }
            }
            else {}
            return component;
        }
        /**
         * Mount component on top of existing html element.
         *
         *     const element = document.createElement("div");
         *     document.body.appendChild(element);
         *     element.innerHTML("<span>content</span>");
         *
         *     const MyComponent = new ComponentDescriptor()
         *       .update((c) => {
         *         c.vSync(c.createVRoot.children([createVElement("span").children("content")]));
         *       });
         *
         *     const component = MyComponent.mountComponent(element);
         *     component.update();
         */
        mountComponent(element, parent, props) {
            const component = new Component(this._markFlags, this, element, parent, props);
            if (this._createState !== null) {
                component.state = this._createState(component, component.props);
            }
            if (this._init !== null) {
                this._init(component, component.props, component.state);
            }
            componentAttached(component);
            return component;
        }
        /**
         * Create event handler.
         */
        createEventHandler(handler) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & 1 /* EnabledBackRef */) === 0) {
                    throw new Error("Failed to create an event handler: component descriptor should have enabled back reference.");
                }
            }
            return function (event) {
                const component = event.currentTarget.xtag;
                if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                    if (component === undefined) {
                        throw new Error(`Failed to dispatch event to event handler: event currentTarget doesn't have back ` +
                            `reference to component.`);
                    }
                }
                handler(event, component, component.props, component.state);
            };
        }
        /**
         * Create delegated event handler.
         */
        createDelegatedEventHandler(selector, componentSelector, handler) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & 1 /* EnabledBackRef */) === 0) {
                    throw new Error("Failed to create an event handler: component descriptor should have enabled back reference.");
                }
            }
            return (event) => {
                let matchingTarget = matchesWithAncestors(event.target, selector, event.currentTarget);
                if (matchingTarget !== null) {
                    let target = matchingTarget;
                    if (typeof componentSelector === "boolean") {
                        if (!componentSelector) {
                            target = event.currentTarget;
                        }
                    }
                    else {
                        target = matchesWithAncestors(matchingTarget, componentSelector, event.currentTarget);
                        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                            if ((this._flags & 1 /* EnabledBackRef */) === 0) {
                                throw new Error(`Failed to dispatch event to event handler: cannot find closest component with ` +
                                    `selector "${componentSelector}".`);
                            }
                        }
                    }
                    const component = target.xtag;
                    if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                        if (component === undefined) {
                            throw new Error(`Failed to dispatch event to event handler: event component target doesn't have back ` +
                                `reference to component.`);
                        }
                    }
                    handler(event, component, component.props, component.state, matchingTarget);
                }
            };
        }
    }
    /**
     * Component.
     *
     * @final
     */
    class Component {
        constructor(flags, descriptor, element, parent, props) {
            this.flags = flags;
            this.mtime = 0;
            this.descriptor = descriptor;
            this.element = element;
            this.depth = parent === undefined ? 0 : parent.depth + 1;
            this.props = props === undefined ? null : props;
            this.state = null;
            this._root = ((flags & 65536 /* Canvas2D */) === 0) ? null : element.getContext("2d");
            this._subscriptions = null;
            this._transientSubscriptions = null;
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                this.element.dxtag = this;
            }
        }
        /**
         * Get canvas 2d rendering context.
         */
        get2DContext() {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this.flags & 65536 /* Canvas2D */) === 0) {
                    throw new Error("Failed to get 2d context: component isn't a canvas.");
                }
            }
            return this._root;
        }
        /**
         * Mark component as dirty and cancel all transient subscriptions.
         */
        markDirty() {
            if ((this.flags & 4 /* Dirty */) === 0) {
                this.flags |= 4 /* Dirty */;
                componentCancelTransientSubscriptions(this);
            }
        }
        /**
         * Creates a virtual dom root node.
         */
        createVRoot() {
            return ((this.flags & 131072 /* VModel */) === 0) ?
                createVRoot() :
                this.descriptor._tag.createVRoot();
        }
        /**
         * **EXPERIMENTAL** Start interaction.
         *
         * When interaction is started, component becomes a high priority target for a scheduler, and scheduler goes into
         * throttled mode.
         */
        startInteraction() {
            if ((this.flags & 512 /* EnabledThrottling */) === 0) {
                this.flags |= 256 /* HighPriorityUpdate */ | 512 /* EnabledThrottling */;
                enableThrottling();
            }
            else {
                this.flags |= 256 /* HighPriorityUpdate */;
            }
        }
        /**
         * **EXPERIMENTAL** Finish interaction.
         *
         * Removes high priority flag from component and disables scheduler throttling.
         */
        finishInteraction() {
            if ((this.flags & 512 /* EnabledThrottling */) === 0) {
                this.flags &= ~256 /* HighPriorityUpdate */;
            }
            else {
                this.flags &= ~(256 /* HighPriorityUpdate */ | 512 /* EnabledThrottling */);
                disableThrottling();
            }
        }
        /**
         * Adds component to a scheduler queue that will update component each animation frame.
         *
         * Component will be updated always, even when scheduler is in throttled mode.
         */
        startUpdateEachFrame() {
            this.flags |= 16 /* UpdateEachFrame */;
            if ((this.flags & 32 /* InUpdateEachFrameQueue */) === 0) {
                this.flags |= 32 /* InUpdateEachFrameQueue */;
                startUpdateComponentEachFrame(this);
            }
        }
        /**
         * Remove component from a scheduler queue that updates component each animation frame.
         */
        stopUpdateEachFrame() {
            this.flags &= ~16 /* UpdateEachFrame */;
        }
        /**
         * Set new props and update component.
         *
         * Props are checked by their identity, unless it is disabled by component descriptor method
         * `disableCheckDataIdentity()`.
         */
        update(newProps) {
            updateComponent(this, newProps);
        }
        /**
         * Sync internal representation using Virtual DOM API.
         *
         * If this method is called during mounting phase, then Virtual DOM will be mounted on top of the existing document
         * tree.
         */
        vSync(newRoot) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((newRoot._flags & 8 /* Root */) === 0) {
                    throw new Error("Failed to sync: sync methods accepts only VNodes representing root node.");
                }
                if ((this.flags & 131072 /* VModel */) !== (newRoot._flags & 131072 /* VModel */)) {
                    if ((this.flags & 131072 /* VModel */) === 0) {
                        throw new Error("Failed to sync: vdom root should have the same type as root registered in component " +
                            "descriptor, component descriptor is using vmodel root.");
                    }
                    else {
                        throw new Error("Failed to sync: vdom root should have the same type as root registered in component " +
                            "descriptor, component descriptor is using simple tag.");
                    }
                }
            }
            if (this._root === null) {
                newRoot.cref = this;
                if ("<@KIVI_MOUNTING@>" === "MOUNTING_ENABLED" && isMounting()) {}
                else {
                    newRoot.ref = this.element;
                    vNodeRender(newRoot, this);
                }
            }
            else {
                syncVNodes(this._root, newRoot, this);
            }
            this._root = newRoot;
        }
        /**
         * Invalidate component and schedule component to update on the next frame.
         *
         * Dirty flags parameter can be used to add hints that describe what has been changed.
         * This metho will automatically cancel all transient subscriptions if preserve transient subscriptions is false.
         */
        invalidate(dirtyFlags = 8 /* DirtyView */, preserveTransientSubscriptions = false) {
            this.flags |= dirtyFlags;
            if ((this.flags & (4 /* Dirty */ | 1 /* Disposed */)) === 0) {
                this.flags |= 4 /* Dirty */;
                if (!preserveTransientSubscriptions) {
                    componentCancelTransientSubscriptions(this);
                }
                nextFrame().updateComponent(this);
            }
        }
        /**
         * Attach method should be invoked when component is attached to the document.
         */
        attach() {
            componentAttached(this);
            if (this._root !== null && ((this.flags & 65536 /* Canvas2D */) === 0)) {
                vNodeAttach(this._root);
            }
        }
        /**
         * Detach method should be invoked when component is detached from the document.
         */
        detach() {
            if (this._root !== null && ((this.flags & 65536 /* Canvas2D */) === 0)) {
                vNodeDetach(this._root);
            }
            componentDetached(this);
        }
        /**
         * Dispose method should be invoked when component is destroyed.
         */
        dispose() {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this.flags & 1 /* Disposed */) !== 0) {
                    throw new Error("Failed to dispose Component: component is already disposed.");
                }
            }
            if ("<@KIVI_COMPONENT_RECYCLING@>" !== "COMPONENT_RECYCLING_ENABLED" ||
                ((this.flags & 524288 /* EnabledRecycling */) === 0) ||
                (this.descriptor._recycledPool.length >= this.descriptor._maxRecycled)) {
                this.flags |= 1 /* Disposed */;
                if (this._root !== null && ((this.flags & 65536 /* Canvas2D */) === 0)) {
                    vNodeDispose(this._root);
                }
                if ((this.flags & 2 /* Attached */) !== 0) {
                    componentDetached(this);
                }
                const disposed = this.descriptor._disposed;
                if (disposed !== null) {
                    disposed(this, this.props, this.state);
                }
            }
            else {}
        }
        /**
         * Subscribe to invalidator object.
         */
        subscribe(invalidator) {
            return invalidator.subscribeComponent(this);
        }
        /**
         * Transiently subscribe to invalidator object.
         *
         * Each time component is invalidated, all transient subscriptions will be canceled.
         */
        transientSubscribe(invalidator) {
            return invalidator.transientSubscribeComponent(this);
        }
    }
    function updateComponent(component, newProps) {
        const flags = component.flags;
        if (newProps !== undefined && (flags & 1024 /* ImmutableProps */) === 0) {
            const oldProps = component.props;
            const newPropsReceived = component.descriptor._newPropsReceived;
            if (newPropsReceived !== null) {
                newPropsReceived(component, oldProps, newProps);
            }
            else {
                component.markDirty();
            }
            component.props = newProps;
        }
        if ((component.flags & (4 /* Dirty */ | 2 /* Attached */)) ===
            (4 /* Dirty */ | 2 /* Attached */)) {
            if (!isThrottled() ||
                isMounting() ||
                ((flags & 256 /* HighPriorityUpdate */) !== 0) ||
                (frameTimeRemaining() > 0)) {
                component.descriptor._update(component, component.props, component.state);
                component.mtime = clock();
                component.flags &= ~(4 /* Dirty */ | 128 /* InUpdateQueue */);
            }
            else {
                nextFrame().updateComponent(component);
            }
        }
    }
    function componentAttached(component) {
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if ((component.flags & 2 /* Attached */) !== 0) {
                throw new Error("Failed to attach Component: component is already attached.");
            }
        }
        component.flags |= 2 /* Attached */;
        if ("<@KIVI_COMPONENT_RECYCLING@>" === "COMPONENT_RECYCLING_ENABLED") {}
        const attached = component.descriptor._attached;
        if (attached !== null) {
            attached(component, component.props, component.state);
        }
    }
    function componentDetached(component) {
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if ((component.flags & 2 /* Attached */) === 0) {
                throw new Error("Failed to detach Component: component is already detached.");
            }
        }
        if ((component.flags & 512 /* EnabledThrottling */) !== 0) {
            disableThrottling();
        }
        component.flags &= ~(2 /* Attached */ | 16 /* UpdateEachFrame */ | 512 /* EnabledThrottling */);
        componentCancelSubscriptions(component);
        componentCancelTransientSubscriptions(component);
        const detached = component.descriptor._detached;
        if (detached !== null) {
            detached(component, component.props, component.state);
        }
    }
    /**
     * Cancel subscriptions.
     */
    function componentCancelSubscriptions(component) {
        let subscription = component._subscriptions;
        while (subscription !== null) {
            subscription._cancel();
            subscription = subscription._componentNext;
        }
        component._subscriptions = null;
    }
    /**
     * Cancel transient subscriptions.
     */
    function componentCancelTransientSubscriptions(component) {
        let subscription = component._transientSubscriptions;
        while (subscription !== null) {
            subscription._cancel();
            subscription = subscription._componentNext;
        }
        component._transientSubscriptions = null;
    }
    /**
     * Inject component into DOM.
     */
    function injectComponent(descriptor, container, props, sync) {
        const c = descriptor.createComponent(undefined, props);
        if (sync) {
            container.appendChild(c.element);
            componentAttached(c);
            updateComponent(c);
        }
        else {
            nextFrame().write(function () {
                container.appendChild(c.element);
                componentAttached(c);
                updateComponent(c);
            });
        }
        return c;
    }
    /**
     * Mount component on top of existing DOM.
     */
    function mountComponent(descriptor, element, props, sync) {
        const c = descriptor.mountComponent(element, undefined, props);
        if (sync) {
            startMounting();
            componentAttached(c);
            updateComponent(c);
            finishMounting();
        }
        else {
            nextFrame().write(function () {
                startMounting();
                componentAttached(c);
                updateComponent(c);
                finishMounting();
            });
        }
        return c;
    }

    /**
     * Minimum time frame duration for throttled tasks.
     */
    const MinThrottleDuration = 2;
    /**
     * Maximum time frame duration for throttled tasks.
     */
    const MaxThrottleDuration = 6;
    /**
     * Default time frame duration for throttled tasks.
     */
    const DefaultThrottleDuration = 6;
    /**
     * Frame tasks group contains tasks for updating components, read dom and write dom tasks, and tasks that should be
     * executed after all other tasks are finished.
     *
     * To get access to the frame tasks group, use: `currentFrame()` and `nextFrame()` scheduler methods.
     *
     *     scheduler.currentFrame().read(() => {
     *       console.log(element.clientWidth);
     *     });
     *
     * @final
     */
    class FrameTasksGroup {
        constructor() {
            this._flags = 0;
            this._componentTasks = [];
            this._writeTasks = null;
            this._readTasks = null;
            this._afterTasks = null;
            this._focus = null;
        }
        /**
         * Add Component to the components queue.
         */
        updateComponent(component) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & 16 /* RWLock */) !== 0) {
                    throw new Error("Failed to add update component task to the current frame, current frame is locked for read" +
                        " and write tasks.");
                }
            }
            if ((component.flags & 128 /* InUpdateQueue */) === 0) {
                component.flags |= 128 /* InUpdateQueue */;
                const priority = component.depth;
                this._flags |= 1 /* Component */;
                while (priority >= this._componentTasks.length) {
                    this._componentTasks.push(null);
                }
                const group = this._componentTasks[priority];
                if (group === null) {
                    this._componentTasks[priority] = [component];
                }
                else {
                    group.push(component);
                }
            }
        }
        /**
         * Add new task to the write DOM task queue.
         */
        write(callback) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & 16 /* RWLock */) !== 0) {
                    throw new Error("Failed to add update component task to the current frame, current frame is locked for read" +
                        " and write tasks.");
                }
            }
            this._flags |= 2 /* Write */;
            if (this._writeTasks === null) {
                this._writeTasks = [];
            }
            this._writeTasks.push(callback);
        }
        /**
         * Add new task to the read DOM task queue.
         */
        read(callback) {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if ((this._flags & 16 /* RWLock */) !== 0) {
                    throw new Error("Failed to add update component task to the current frame, current frame is locked for read" +
                        " and write tasks.");
                }
            }
            this._flags |= 4 /* Read */;
            if (this._readTasks === null) {
                this._readTasks = [];
            }
            this._readTasks.push(callback);
        }
        /**
         * Add new task to the task queue that will execute tasks when all DOM tasks are finished.
         */
        after(callback) {
            this._flags |= 8 /* After */;
            if (this._afterTasks === null) {
                this._afterTasks = [];
            }
            this._afterTasks.push(callback);
        }
        /**
         * Set a focus on an element when all DOM tasks are finished.
         */
        focus(node) {
            this._focus = node;
        }
        /**
         * Place a lock on adding new read and write task.
         *
         * Works in DEBUG mode only.
         */
        _rwLock() {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                this._flags |= 16 /* RWLock */;
            }
        }
        /**
         * Remove a lock from adding new read and write tasks.
         *
         * Works in DEBUG mode only.
         */
        _rwUnlock() {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                this._flags &= ~16 /* RWLock */;
            }
        }
    }
    /**
     * Global scheduler variables.
     *
     * Scheduler supports animation frame tasks, macrotasks and microtasks.
     *
     * Animation frame tasks will be executed in batches, switching between write and read tasks until there
     * are no tasks left. Write tasks are sorted by their priority, tasks with the lowest priority value are
     * executed first, so the lowest depth of the Component in the components tree has highest priority.
     *
     * Scheduler also have monotonically increasing internal clock, it increments each time scheduler goes
     * from one animation frame to another, or starts executing macrotasks or microtasks.
     */
    const scheduler = {
        /**
         * See `SchedulerFlags` for details.
         */
        flags: 0,
        clock: 0,
        time: 0,
        microtasks: [],
        macrotasks: [],
        currentFrame: new FrameTasksGroup(),
        nextFrame: new FrameTasksGroup(),
        updateComponents: [],
        microtaskNode: document.createTextNode(""),
        microtaskToggle: 0,
        macrotaskMessage: "__kivi" + Math.random(),
        throttleEnabledCounter: 0,
        throttledFrameDuration: DefaultThrottleDuration,
        throttledFps: 60,
        throttledDiffWindow: 0,
        throttledFrameDeadline: 0,
    };
    // Microtask scheduler based on mutation observer
    const microtaskObserver = new MutationObserver(runMicrotasks);
    microtaskObserver.observe(scheduler.microtaskNode, { characterData: true });
    // Macrotask scheduler based on postMessage
    window.addEventListener("message", handleWindowMessage);
    scheduler.currentFrame._rwLock();
    /**
     * Returns current monotonically increasing clock.
     */
    function clock() {
        return scheduler.clock;
    }
    function requestMicrotaskExecution() {
        if ((scheduler.flags & 1 /* MicrotaskPending */) === 0) {
            scheduler.flags |= 1 /* MicrotaskPending */;
            scheduler.microtaskToggle ^= 1;
            scheduler.microtaskNode.nodeValue = scheduler.microtaskToggle ? "1" : "0";
        }
    }
    function requestMacrotaskExecution() {
        if ((scheduler.flags & 2 /* MacrotaskPending */) === 0) {
            scheduler.flags |= 2 /* MacrotaskPending */;
            window.postMessage(scheduler.macrotaskMessage, "*");
        }
    }
    function requestNextFrame() {
        if ((scheduler.flags & 4 /* FrametaskPending */) === 0) {
            scheduler.flags |= 4 /* FrametaskPending */;
            requestAnimationFrame(handleNextFrame);
        }
    }
    function handleWindowMessage(e) {
        if (e.source === window && e.data === scheduler.macrotaskMessage) {
            runMacrotasks();
        }
    }
    function handleNextFrame(t) {
        const updateComponents = scheduler.updateComponents;
        let tasks;
        let i;
        let j;
        scheduler.flags &= ~(4 /* FrametaskPending */ | 16 /* ThrottledFrameExhausted */);
        scheduler.time = Date.now();
        if ((scheduler.flags & 8 /* EnabledThrottling */) !== 0) {
            scheduleMacrotask(() => {
                const elapsed = (window.performance.now() - t) / 1000;
                scheduler.throttledFps = Math.round((scheduler.throttledFps + (1 / elapsed)) / 2);
                scheduler.throttledDiffWindow += (scheduler.throttledFps < 45) ? -1 : 1;
                if (scheduler.throttledDiffWindow > 5) {
                    scheduler.throttledDiffWindow = 0;
                    scheduler.throttledFrameDuration += 0.1;
                }
                else if (scheduler.throttledDiffWindow < -5) {
                    scheduler.throttledDiffWindow = 0;
                    scheduler.throttledFrameDuration *= 0.66;
                }
                if (scheduler.throttledFrameDuration > MaxThrottleDuration) {
                    scheduler.throttledFrameDuration = MaxThrottleDuration;
                }
                else if (scheduler.throttledFrameDuration < MinThrottleDuration) {
                    scheduler.throttledFrameDuration = MinThrottleDuration;
                }
            });
            scheduler.throttledFrameDeadline = t + scheduler.throttledFrameDuration;
        }
        const frame = scheduler.nextFrame;
        scheduler.nextFrame = scheduler.currentFrame;
        scheduler.currentFrame = frame;
        scheduler.currentFrame._rwUnlock();
        scheduler.nextFrame._rwUnlock();
        // Mark all update components as dirty. But don't update until all write tasks are finished. It is possible that we
        // won't need to update component if it is removed.
        for (i = 0; i < updateComponents.length; i++) {
            updateComponents[i].flags |= 4 /* Dirty */;
        }
        // Perform read/write batching. Start with executing read DOM tasks, then update components, execute write DOM tasks
        // and repeat until all read and write tasks are executed.
        do {
            while ((frame._flags & 4 /* Read */) !== 0) {
                frame._flags &= ~4 /* Read */;
                tasks = frame._readTasks;
                frame._readTasks = null;
                for (i = 0; i < tasks.length; i++) {
                    tasks[i]();
                }
            }
            while ((frame._flags & (1 /* Component */ | 2 /* Write */)) !== 0) {
                if ((frame._flags & 1 /* Component */) !== 0) {
                    frame._flags &= ~1 /* Component */;
                    const componentGroups = frame._componentTasks;
                    for (i = 0; i < componentGroups.length; i++) {
                        const componentGroup = componentGroups[i];
                        if (componentGroup !== null) {
                            componentGroups[i] = null;
                            for (j = 0; j < componentGroup.length; j++) {
                                updateComponent(componentGroup[j]);
                            }
                        }
                    }
                }
                if ((frame._flags & 2 /* Write */) !== 0) {
                    frame._flags &= ~2 /* Write */;
                    tasks = frame._writeTasks;
                    frame._writeTasks = null;
                    for (i = 0; i < tasks.length; i++) {
                        tasks[i]();
                    }
                }
            }
            // Update components registered for updating on each frame.
            // Remove components that doesn't have UPDATE_EACH_FRAME flag.
            i = 0;
            j = updateComponents.length;
            while (i < j) {
                const component = updateComponents[i++];
                if ((component.flags & 16 /* UpdateEachFrame */) === 0) {
                    component.flags &= ~32 /* InUpdateEachFrameQueue */;
                    if (i === j) {
                        updateComponents.pop();
                    }
                    else {
                        updateComponents[--i] = updateComponents.pop();
                    }
                }
                else {
                    updateComponent(component);
                }
            }
        } while ((frame._flags & (1 /* Component */ |
            2 /* Write */ |
            4 /* Read */)) !== 0);
        // Lock current from adding read and write tasks in debug mode.
        scheduler.currentFrame._rwLock();
        // Perform tasks that should be executed when all DOM ops are finished.
        while ((frame._flags & 8 /* After */) !== 0) {
            frame._flags &= ~8 /* After */;
            tasks = frame._afterTasks;
            frame._afterTasks = null;
            for (i = 0; i < tasks.length; i++) {
                tasks[i]();
            }
        }
        // Set focus on an element.
        if (frame._focus !== null) {
            if (frame._focus.constructor === VNode) {
                frame._focus.ref.focus();
            }
            else {
                frame._focus.focus();
            }
            frame._focus = null;
        }
        if (updateComponents.length > 0) {
            requestNextFrame();
        }
        scheduler.clock++;
    }
    function runMicrotasks() {
        scheduler.time = Date.now();
        while (scheduler.microtasks.length > 0) {
            const tasks = scheduler.microtasks;
            scheduler.microtasks = [];
            for (let i = 0; i < tasks.length; i++) {
                tasks[i]();
            }
            scheduler.clock++;
        }
        scheduler.flags &= ~1 /* MicrotaskPending */;
    }
    function runMacrotasks() {
        scheduler.flags &= ~2 /* MacrotaskPending */;
        scheduler.time = Date.now();
        let tasks = scheduler.macrotasks;
        scheduler.macrotasks = [];
        for (let i = 0; i < tasks.length; i++) {
            tasks[i]();
        }
        scheduler.clock++;
    }
    /**
     * Add task to the microtask queue.
     */
    function scheduleMicrotask(task) {
        requestMicrotaskExecution();
        scheduler.microtasks.push(task);
    }
    /**
     * Add task to the macrotask queue.
     */
    function scheduleMacrotask(task) {
        requestMacrotaskExecution();
        scheduler.macrotasks.push(task);
    }
    /**
     * Get task list for the current frame.
     */
    function currentFrame() {
        return scheduler.currentFrame;
    }
    /**
     * Get task list for the next frame.
     */
    function nextFrame() {
        requestNextFrame();
        return scheduler.nextFrame;
    }
    /**
     * **EXPERIMENTAL** Enable throttling mode.
     *
     * In throttling mode, all updates for low priority components will be throttled, all updates will be performed
     * incrementally each frame.
     *
     * Each time throttling is enabled, it increases internal counter that tracks how many times it is enabled, and when
     * counters goes to zero, throttling mode will be disabled.
     */
    function enableThrottling() {
        scheduler.throttleEnabledCounter++;
        scheduler.flags |= 8 /* EnabledThrottling */;
    }
    /**
     * **EXPERIMENTAL** Disable throttling mode.
     *
     * Throttling mode will be disabled when number of dependencies that enabled throttling mode goes to zero.
     */
    function disableThrottling() {
        if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
            if (scheduler.throttleEnabledCounter < 0) {
                throw new Error("Failed to disable scheduler throttling, it is already disabled.");
            }
        }
        scheduler.throttleEnabledCounter--;
        if (scheduler.throttleEnabledCounter === 0) {
            scheduler.flags &= ~8 /* EnabledThrottling */;
        }
    }
    /**
     * Get remaining time available for tasks in the current frame.
     */
    function frameTimeRemaining() {
        if ((scheduler.flags & 16 /* ThrottledFrameExhausted */) !== 0) {
            return 0;
        }
        else {
            const remaining = scheduler.throttledFrameDeadline - performance.now();
            if (remaining <= 0) {
                scheduler.flags |= 16 /* ThrottledFrameExhausted */;
                return 0;
            }
            else {
                return remaining;
            }
        }
    }
    /**
     * Add component to the list of components that should be updated each frame.
     */
    function startUpdateComponentEachFrame(component) {
        requestNextFrame();
        scheduler.updateComponents.push(component);
    }
    /**
     * Returns true if UI state is in mounting phase.
     */
    function isMounting() {
        return ((scheduler.flags & 32 /* EnabledMounting */) !== 0);
    }
    /**
     * Returns true if tasks should be throttled.
     */
    function isThrottled() {
        return ((scheduler.flags & 8 /* EnabledThrottling */) !== 0);
    }
    /**
     * Start UI mounting phase.
     */
    function startMounting() {
        scheduler.flags |= 32 /* EnabledMounting */;
    }
    /**
     * Finish UI mounting phase.
     */
    function finishMounting() {
        scheduler.flags &= ~32 /* EnabledMounting */;
    }

    /**
     * Model for DOM Elements.
     *
     * Models are used as an advanced optimization technique. When creating virtual nodes, or declaring root node for a
     * component, it is possible to link them to a model instead of simple HTML tagName. Model will contain all static
     * properties for the HTML Element, so there is no need to declare them each time virtual node is created. It also
     * reduces diff overhead, because there is no need to diff static model properties.
     *
     * Creating a virtual dom node from VModel:
     *
     *     const model = new VModel("div").attrs({"id": "model"});
     *     // Element node
     *     const node = mode.createVNode();
     *     // Component's root node
     *     const root = mode.createVRoot();
     *
     * Creating a DOM node from VModel:
     *
     *     const model = new VModel("div").attrs({"id": "model"});
     *     const div = model.createElement();
     *
     * @final
     */
    class VModel {
        constructor(tagName) {
            this._markFlags = 131072 /* VModel */;
            this._flags = 0;
            this._tagName = tagName;
            this._props = null;
            this._attrs = null;
            this._style = null;
            this._className = null;
            this._updateHandler = null;
            this._ref = null;
        }
        /**
         * Use svg namespace for the DOM element.
         */
        svg() {
            this._markFlags |= 32768 /* Svg */;
            this._flags |= 32768 /* Svg */;
            return this;
        }
        /**
         * Set properties.
         *
         * All properties are assigned to DOM nodes directly:
         *
         *     e: HTMLElement;
         *     e.propertyName = propertyValue;
         *
         * When virtual node is mounted on top of existing HTML, all properties from model will be assigned during mounting
         * phase.
         */
        props(props) {
            this._props = props;
            return this;
        }
        /**
         * Set attributes.
         *
         * All attributes are assigned to DOM nodes with `setAttribute` method:
         *
         *     e: HTMLElement;
         *     e.setAttribute(key, value);
         *
         * If attribute is prefixed with "xlink:", or "xml:" namespace, it will assign attributes with `setAttributeNS`
         * method and use appropriate namespaces.
         */
        attrs(attrs) {
            this._attrs = attrs;
            return this;
        }
        /**
         * Set style in css string format.
         *
         * Style is assigned to DOM nodes with `style.cssText` property, if virtual node represents an element from svg
         * namespace, style will be assigned with `setAttribute("style", "cssText")` method.
         */
        style(style) {
            this._style = style;
            return this;
        }
        /**
         * Set className.
         *
         * Class name is assigned to DOM nodes with `className` property, if virtual node represents an element from svg
         * namespace, class name will be assigned with `setAttribute("class", "className")` method.
         */
        className(classes) {
            this._className = classes;
            return this;
        }
        /**
         * Enable node cloning.
         *
         * Instead of creating DOM nodes, model will clone nodes from a base node with `Node.cloneNode(false)` method.
         */
        enableCloning() {
            this._flags |= 1 /* EnabledCloning */;
            return this;
        }
        /**
         * Set update handler.
         *
         * Update handler is used to override default reconciliation algorithm.
         */
        updateHandler(handler) {
            this._markFlags |= 262144 /* VModelUpdateHandler */;
            this._updateHandler = handler;
            return this;
        }
        /**
         * Create a Virtual DOM Node.
         */
        createVNode(data) {
            return new VNode(2 /* Element */ | this._markFlags, this, data === undefined ? null : data);
        }
        /**
         * Create a Virtual DOM Node for component's root.
         */
        createVRoot(data) {
            return new VNode(8 /* Root */ | this._markFlags, this, data === undefined ? null : data);
        }
        /**
         * Create a DOM Element.
         */
        createElement() {
            let i;
            let keys;
            let key;
            let ref = this._ref;
            if (ref === null) {
                if ((this._flags & 32768 /* Svg */) === 0) {
                    ref = document.createElement(this._tagName);
                }
                else {
                    ref = document.createElementNS(SvgNamespace, this._tagName);
                }
                if (this._props !== null) {
                    keys = Object.keys(this._props);
                    for (i = 0; i < keys.length; i++) {
                        key = keys[i];
                        ref[key] = this._props[key];
                    }
                }
                if (this._attrs !== null) {
                    keys = Object.keys(this._attrs);
                    for (i = 0; i < keys.length; i++) {
                        key = keys[i];
                        setAttr(ref, key, this._attrs[key]);
                    }
                }
                if (this._style !== null) {
                    if ((this._flags & 32768 /* Svg */) === 0) {
                        ref.style.cssText = this._style;
                    }
                    else {
                        ref.setAttribute("style", this._style);
                    }
                }
                if (this._className !== null) {
                    if ((this._flags & 32768 /* Svg */) === 0) {
                        ref.className = this._className;
                    }
                    else {
                        ref.setAttribute("class", this._className);
                    }
                }
                if ((this._flags & 1 /* EnabledCloning */) !== 0) {
                    this._ref = ref;
                    return ref.cloneNode(false);
                }
                return ref;
            }
            else {
                return ref.cloneNode(false);
            }
        }
        /**
         * Update DOM Node with an update handler.
         */
        update(element, oldProps, newProps) {
            this._updateHandler(element, oldProps, newProps);
        }
    }

    /**
     * Invalidator Subscription.
     *
     * @final
     */
    class InvalidatorSubscription {
        constructor(flags, invalidator, component, callback) {
            this._flags = flags;
            this.invalidator = invalidator;
            this._component = component;
            this._callback = callback;
            this._invalidatorPrev = null;
            this._invalidatorNext = null;
            this._componentPrev = null;
            this._componentNext = null;
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                this._isCanceled = false;
            }
        }
        _cancel() {
            if (this._invalidatorPrev === null) {
                if ((this._flags & 2 /* Transient */) === 0) {
                    this.invalidator._subscriptions = this._invalidatorNext;
                }
                else {
                    this.invalidator._transientSubscriptions = this._invalidatorNext;
                }
            }
            else {
                this._invalidatorPrev._invalidatorNext = this._invalidatorNext;
            }
            if (this._invalidatorNext !== null) {
                this._invalidatorNext._invalidatorPrev = this._invalidatorPrev;
            }
        }
        /**
         * Cancel subscription.
         */
        cancel() {
            if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
                if (this._isCanceled) {
                    throw new Error("Failed to cancel InvalidatorSubscription: subscription cannot be canceled twice.");
                }
                this._isCanceled = true;
            }
            this._cancel();
            if ((this._flags & 1 /* Component */) !== 0) {
                if (this._componentPrev === null) {
                    if ((this._flags & 2 /* Transient */) === 0) {
                        this._component._subscriptions = this._componentNext;
                    }
                    else {
                        this._component._transientSubscriptions = this._componentNext;
                    }
                }
                else {
                    this._componentPrev._componentNext = this._componentNext;
                }
                if (this._componentNext !== null) {
                    this._componentNext._componentPrev = this._componentPrev;
                }
            }
        }
        _invalidate() {
            if ((this._flags & 1 /* Component */) !== 0) {
                this._component.invalidate();
            }
            else {
                this._callback();
            }
        }
    }
    /**
     * Invalidator object is a low level object for implementing reactive bindings.
     *
     * @final
     */
    class Invalidator {
        constructor() {
            this.mtime = clock();
            this._subscriptions = null;
            this._transientSubscriptions = null;
        }
        /**
         * Create a subscription with a callback handler.
         */
        subscribe(callback) {
            const s = new InvalidatorSubscription(0, this, null, callback);
            const firstSubscription = this._subscriptions;
            if (firstSubscription !== null) {
                firstSubscription._invalidatorPrev = s;
                s._invalidatorNext = firstSubscription;
            }
            this._subscriptions = s;
            return s;
        }
        /**
         * Create a transient subscription with a callback handler.
         */
        transientSubscribe(callback) {
            const s = new InvalidatorSubscription(2 /* Transient */, this, null, callback);
            const firstSubscription = this._transientSubscriptions;
            if (firstSubscription !== null) {
                firstSubscription._invalidatorPrev = s;
                s._invalidatorNext = firstSubscription;
            }
            this._transientSubscriptions = s;
            return s;
        }
        /**
         * Create a subscription that will invalidate component.
         */
        subscribeComponent(component) {
            const s = new InvalidatorSubscription(1 /* Component */, this, component, null);
            const firstSubscription = this._subscriptions;
            const firstComponentSubscription = component._subscriptions;
            if (firstSubscription !== null) {
                firstSubscription._invalidatorPrev = s;
                s._invalidatorNext = firstSubscription;
            }
            if (firstComponentSubscription !== null) {
                firstComponentSubscription._componentPrev = s;
                s._componentNext = firstComponentSubscription;
            }
            this._subscriptions = s;
            component._subscriptions = s;
            return s;
        }
        /**
         * Create a transient subscription that will invalidate component.
         */
        transientSubscribeComponent(component) {
            const s = new InvalidatorSubscription(1 /* Component */ | 2 /* Transient */, this, component, null);
            const firstSubscription = this._transientSubscriptions;
            const firstComponentSubscription = component._transientSubscriptions;
            if (firstSubscription !== null) {
                firstSubscription._invalidatorPrev = s;
                s._invalidatorNext = firstSubscription;
            }
            if (firstComponentSubscription !== null) {
                firstComponentSubscription._componentPrev = s;
                s._componentNext = firstComponentSubscription;
            }
            this._transientSubscriptions = s;
            component._transientSubscriptions = s;
            return s;
        }
        /**
         * Returns true if invalidator object has subscriptions.
         */
        hasSubscriptions() {
            return (this._subscriptions !== null || this._transientSubscriptions !== null);
        }
        /**
         * Invalidate all subscriptions.
         */
        invalidate() {
            const now = clock();
            if (this.mtime < now) {
                this.mtime = now;
                let subscription = this._subscriptions;
                while (subscription !== null) {
                    subscription._invalidate();
                    subscription = subscription._invalidatorNext;
                }
                subscription = this._transientSubscriptions;
                while (subscription !== null) {
                    subscription._invalidate();
                    subscription = subscription._invalidatorNext;
                }
                this._transientSubscriptions = null;
            }
        }
    }

    /**
     * Recursively flattens VNode arrays and skips null nodes.
     */
    function normalizeVNodes(nodes) {
        let copy = nodes.slice(0);
        const flatten = [];
        while (copy.length > 0) {
            const item = copy.shift();
            if (item !== null) {
                if (item.constructor === VNode) {
                    flatten.push(item);
                }
                else {
                    copy = item.concat(copy);
                }
            }
        }
        return flatten;
    }

    if ("<@KIVI_DEBUG@>" !== "DEBUG_DISABLED") {
        console.info("KIVI_DEBUG: enabled");
        if ("<@KIVI_COMPONENT_RECYCLING@>" === "COMPONENT_RECYCLING_ENABLED") {}
        else {
            console.info("KIVI_COMPONENT_RECYCLING: disabled");
        }
    }

    exports.clock = clock;
    exports.scheduleMicrotask = scheduleMicrotask;
    exports.scheduleMacrotask = scheduleMacrotask;
    exports.currentFrame = currentFrame;
    exports.nextFrame = nextFrame;
    exports.enableThrottling = enableThrottling;
    exports.disableThrottling = disableThrottling;
    exports.isMounting = isMounting;
    exports.VModel = VModel;
    exports.InvalidatorSubscription = InvalidatorSubscription;
    exports.Invalidator = Invalidator;
    exports.VNode = VNode;
    exports.createVText = createVText;
    exports.createVElement = createVElement;
    exports.createVSvgElement = createVSvgElement;
    exports.createVRoot = createVRoot;
    exports.ComponentDescriptorRegistry = ComponentDescriptorRegistry;
    exports.ComponentDescriptor = ComponentDescriptor;
    exports.Component = Component;
    exports.injectComponent = injectComponent;
    exports.mountComponent = mountComponent;
    exports.getTagName = getTagName;
    exports.getClassName = getClassName;
    exports.matchesWithAncestors = matchesWithAncestors;
    exports.normalizeVNodes = normalizeVNodes;

    Object.defineProperty(exports, '__esModule', { value: true });

}));